# Copyright (c) 2025 JP Hutchins
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Any, Final, NamedTuple, assert_type

import pytest

from mahonia import (
	Add,
	And,
	Approximately,
	BoundExpr,
	Const,
	Eq,
	Expr,
	Lt,
	Not,
	Percent,
	PlusMinus,
	Pow,
	Predicate,
	TSupportsComparison,
	Var,
)


@dataclass(frozen=True)
class Ctx:
	x: int
	y: int
	name: str
	flag: bool = True
	custom: object = object()
	f: float = 1.5
	e: float = 2.71828


ctx: Final = Ctx(x=5, y=10, name="example")


def between(
	expr: Expr[TSupportsComparison, Ctx], low: TSupportsComparison, high: TSupportsComparison
) -> "And[bool, Ctx]":
	"""Example of defining some convenience to compose an expression."""
	return And(  # type: ignore[arg-type]
		Lt(Const("Low", low), expr),  # type: ignore[arg-type]
		Lt(expr, Const("High", high)),  # type: ignore[arg-type]
	)


def test_const_name_eval_and_str() -> None:
	c = Const("Forty-two", 42)
	assert c.value == 42
	assert c.unwrap(ctx) == 42
	assert c.to_string() == "Forty-two:42"
	assert c.to_string(ctx) == "Forty-two:42"


def test_const_eval_and_str() -> None:
	c = Const(None, 100)
	assert c.value == 100
	assert c.unwrap(ctx) == 100
	assert c.to_string() == "100"
	assert c.to_string(ctx) == "100"


@pytest.mark.mypy_testing
def test_const_generic_type() -> None:
	c_int = Const(None, 1)
	assert_type(c_int, Const[int])
	assert_type(c_int.value, int)

	c_str = Const(None, "test")
	assert_type(c_str, Const[str])
	assert_type(c_str.value, str)

	c_float = Const(None, 1.5)
	assert_type(c_float, Const[float])
	assert_type(c_float.value, float)

	c_complex = Const(None, 1 + 2j)
	assert_type(c_complex, Const[complex])
	assert_type(c_complex.value, complex)

	c_bool = Const(None, True)
	assert_type(c_bool, Const[bool])
	assert_type(c_bool.value, bool)

	class CustomType:
		pass

	c_custom = Const(None, CustomType())
	assert_type(c_custom, Const[CustomType])
	assert_type(c_custom.value, CustomType)


@pytest.mark.mypy_testing
def test_var_generic_type() -> None:
	v_int = Var[int, Ctx]("x")
	assert_type(v_int, Var[int, Ctx])
	assert_type(v_int.unwrap(ctx), int)

	v_str = Var[str, Ctx]("name")
	assert_type(v_str, Var[str, Ctx])
	assert_type(v_str.unwrap(ctx), str)

	v_float = Var[float, Ctx]("y")
	assert_type(v_float, Var[float, Ctx])
	assert_type(v_float.unwrap(ctx), float)

	v_bool = Var[bool, Ctx]("flag")
	assert_type(v_bool, Var[bool, Ctx])
	assert_type(v_bool.unwrap(ctx), bool)

	class CustomType:
		pass

	v_custom = Var[CustomType, Ctx]("custom")
	assert_type(v_custom, Var[CustomType, Ctx])
	assert_type(v_custom.unwrap(ctx), CustomType)


@pytest.mark.mypy_testing
def test_eq_generic_type() -> None:
	v_int = Var[int, Ctx]("x")
	c_int = Const("Five", 5)
	eq_expr = v_int == c_int
	assert_type(eq_expr, Eq[int, Ctx])
	assert_type(eq_expr.unwrap(ctx), bool)

	v_str = Var[str, Ctx]("name")
	c_str = Const("Example", "example")
	eq_expr_str = v_str == c_str
	assert_type(eq_expr_str, Eq[str, Ctx])
	assert_type(eq_expr_str.unwrap(ctx), bool)


@pytest.mark.mypy_testing
def test_add_generic_type() -> None:
	v_int = Var[int, Ctx]("x")
	c_int = Const("Five", 5)
	add_expr = v_int + c_int
	assert_type(add_expr, Add[int, Ctx])
	assert_type(add_expr.unwrap(ctx), int)

	v_float = Var[float, Ctx]("y")
	c_float = Const("Two", 2.0)
	add_expr_float = v_float + c_float
	assert_type(add_expr_float, Add[float, Ctx])
	assert_type(add_expr_float.unwrap(ctx), float)


def test_var_eval_and_str() -> None:
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	name = Var[str, Ctx]("name")
	assert x.unwrap(ctx) == 5
	assert y.unwrap(ctx) == 10
	assert name.unwrap(ctx) == "example"
	assert x.to_string() == "x"
	assert x.to_string(ctx) == "x:5"
	assert name.to_string(ctx) == "name:example"


def test_add_sub_mul_div() -> None:
	x = Var[float, Ctx]("x")
	y = Var[float, Ctx]("y")
	c5 = Const("Five", 5.0)
	c2 = Const("Two", 2.0)
	assert (x + y).unwrap(ctx) == 15
	assert (x + c5).unwrap(ctx) == 10
	assert (y - x).unwrap(ctx) == 5
	assert (x * c2).unwrap(ctx) == 10
	assert (y / c2).unwrap(ctx) == 5
	assert (x + y).to_string() == "(x + y)"
	assert (x + c5).to_string(ctx) == "(x:5 + Five:5.0 -> 10.0)"


def test_add() -> None:
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	assert (x + y).unwrap(ctx) == 15
	assert (x + 5).unwrap(ctx) == 10
	assert (x + y).to_string() == "(x + y)"
	assert (x + 5).to_string(ctx) == "(x:5 + 5 -> 10)"


def test_comparisons() -> None:
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	c5 = Const("Five", 5)
	c10 = Const("Ten", 10)
	assert (x == c5).unwrap(ctx) is True
	assert (y == c5).unwrap(ctx) is False
	assert (x != c5).unwrap(ctx) is False
	assert (y != c5).unwrap(ctx) is True
	assert (x < y).unwrap(ctx) is True
	assert (y > x).unwrap(ctx) is True
	assert (x <= c5).unwrap(ctx) is True
	assert (y >= c10).unwrap(ctx) is True
	assert (x == c5).to_string() == "(x == Five:5)"
	assert (x == c5).to_string(ctx) == "(x:5 == Five:5 -> True)"


def test_logical_ops() -> None:
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	c5 = Const("Five", 5)
	c10 = Const("Ten", 10)
	pred = ((x == c5) & (y == c10)) | (x != c5)
	assert pred.unwrap(ctx) is True
	assert pred.to_string() == "(((x == Five:5) & (y == Ten:10)) | (x != Five:5))"  # noqa: E501
	# Evaluate with context
	assert (
		pred.to_string(ctx)
		== "(((x:5 == Five:5 -> True) & (y:10 == Ten:10 -> True) -> True) | (x:5 != Five:5 -> False) -> True)"  # noqa: E501
	)


def test_not() -> None:
	x = Var[int, Ctx]("x")
	c5 = Const("name", 5)
	expr = ~(x == c5)
	assert expr.unwrap(ctx) is False
	assert expr.to_string() == "(not (x == name:5))"
	assert expr.to_string(ctx) == "(not (x:5 == name:5 -> True) -> False)"


def test_nested_arithmetic() -> None:
	x = Var[float, Ctx]("x")
	y = Var[float, Ctx]("y")
	expr = (x * Const("name", 2.0)) + (y / Const("name", 2.0))
	assert expr.unwrap(ctx) == 15
	assert expr.to_string() == "((x * name:2.0) + (y / name:2.0))"
	assert expr.to_string(ctx) == "((x:5 * name:2.0 -> 10.0) + (y:10 / name:2.0 -> 5.0) -> 15.0)"


def test_constants_only() -> None:
	c10 = Const("name", 10.0)
	c5 = Const("name", 5.0)
	assert (c10 + c5).unwrap(None) == 15
	assert (c10 / c5).unwrap(None) == 2
	assert (c10 == c10).unwrap(None) is True
	assert (c10 == c5).unwrap(None) is False
	assert (c10 == 10.0).unwrap(None) is True
	assert (c10 != 5.0).unwrap(None) is True
	assert (c10 > c5).unwrap(None) is True
	assert (c10 + c5 * 50).unwrap(None) == 260.0
	assert (c10 + c5 * 50).to_string() == "(name:10.0 + (name:5.0 * 50))"
	assert (c10 + c5 * 50).to_string(ctx) == "(name:10.0 + (name:5.0 * 50 -> 250.0) -> 260.0)"


def test_chained_arithmetic() -> None:
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = x + y * 2 - 3
	assert expr.unwrap(ctx) == 5 + 10 * 2 - 3


def test_const_to_string_edge_cases() -> None:
	c_none = Const(None, None)
	assert c_none.to_string() == "None"
	assert c_none.unwrap(ctx) is None


def test_bool_logic() -> None:
	flag = Var[bool, Ctx]("flag")
	c_true = Const("True", True)
	c_false = Const("False", False)
	assert (flag & c_true).unwrap(ctx) is True
	assert (flag | c_false).unwrap(ctx) is True
	assert (~flag).unwrap(ctx) is False


def test_const_vs_python_literal() -> None:
	c10 = Const("name", 10.0)
	expr = c10 == 10.0
	assert expr.unwrap(None) is True
	assert (c10 != 5.0).unwrap(None) is True
	assert (c10 > 5.0).unwrap(None) is True
	assert (c10 < 20.0).unwrap(None) is True


def test_var_add_const_and_literal() -> None:
	x = Var[int, Ctx]("x")
	c5 = Const("Five", 5)
	assert (x + c5).unwrap(ctx) == 10
	assert (x + 2).unwrap(ctx) == 7


def test_const_add_var_and_literal() -> None:
	x = Var[int, Ctx]("x")
	c5 = Const("Five", 5)
	assert (c5 + x).unwrap(ctx) == 10
	assert (c5 + 2).unwrap(ctx) == 7


def test_var_sub_const_and_literal() -> None:
	x = Var[int, Ctx]("x")
	c5 = Const("Five", 5)
	assert (x - c5).unwrap(ctx) == 0
	assert (x - 2).unwrap(ctx) == 3


def test_const_sub_var_and_literal() -> None:
	x = Var[int, Ctx]("x")
	c5 = Const("Five", 5)
	assert (c5 - x).unwrap(ctx) == 0
	assert (c5 - 2).unwrap(ctx) == 3


def test_var_mul_const_and_literal() -> None:
	x = Var[int, Ctx]("x")
	c5 = Const("Five", 5)
	assert (x * c5).unwrap(ctx) == 25
	assert (x * 2).unwrap(ctx) == 10


def test_const_mul_var_and_literal() -> None:
	x = Var[int, Ctx]("x")
	c5 = Const("Five", 5)
	assert (c5 * x).unwrap(ctx) == 25
	assert (c5 * 2).unwrap(ctx) == 10


def test_var_truediv_const_and_literal() -> None:
	y = Var[float, Ctx]("y")
	c5 = Const("Five", 5.0)
	assert (y / c5).unwrap(ctx) == 2
	assert (y / 2.0).unwrap(ctx) == 5


def test_const_truediv_var_and_literal() -> None:
	y = Var[float, Ctx]("y")
	c10 = Const("Ten", 10)
	assert (c10 / y).unwrap(ctx) == 1
	assert (c10 / 2).unwrap(ctx) == 5


def test_var_eq_const() -> None:
	x = Var[int, Ctx]("x")
	c5 = Const("Five", 5)
	assert (x == c5).unwrap(ctx) is True


def test_const_eq_var() -> None:
	x = Var[int, Ctx]("x")
	c5 = Const("Five", 5)
	assert (c5 == x).unwrap(ctx) is True


def test_var_ne_const_and_literal() -> None:
	x = Var[int, Ctx]("x")
	c5 = Const("Five", 5)
	assert (x != c5).unwrap(ctx) is False
	assert (x != 7).unwrap(ctx) is True


def test_const_ne_var_and_literal() -> None:
	x = Var[int, Ctx]("x")
	c5 = Const("Five", 5)
	assert (c5 != x).unwrap(ctx) is False
	assert (c5 != 7).unwrap(ctx) is True


def test_var_lt_const_and_literal() -> None:
	x = Var[int, Ctx]("x")
	c10 = Const("Ten", 10)
	assert (x < c10).unwrap(ctx) is True
	assert (x < 3).unwrap(ctx) is False


def test_const_lt_var_and_literal() -> None:
	x = Var[int, Ctx]("x")
	c3 = Const("Three", 3)
	assert (c3 < x).unwrap(ctx) is True
	assert (c3 < 2).unwrap(ctx) is False


def test_var_le_const_and_literal() -> None:
	x = Var[int, Ctx]("x")
	c5 = Const("Five", 5)
	assert (x <= c5).unwrap(ctx) is True
	assert (x <= 4).unwrap(ctx) is False


def test_const_le_var_and_literal() -> None:
	x = Var[int, Ctx]("x")
	c5 = Const("Five", 5)
	assert (c5 <= x).unwrap(ctx) is True
	assert (c5 <= 4).unwrap(ctx) is False


def test_var_gt_const_and_literal() -> None:
	y = Var[int, Ctx]("y")
	c5 = Const("Five", 5)
	assert (y > c5).unwrap(ctx) is True
	assert (y > 20).unwrap(ctx) is False


def test_const_gt_var_and_literal() -> None:
	y = Var[int, Ctx]("y")
	c20 = Const("Twenty", 20)
	assert (c20 > y).unwrap(ctx) is True
	assert (c20 > 30).unwrap(ctx) is False


def test_var_ge_const_and_literal() -> None:
	y = Var[int, Ctx]("y")
	c10 = Const("Ten", 10)
	assert (y >= c10).unwrap(ctx) is True
	assert (y >= 20).unwrap(ctx) is False


def test_const_ge_var_and_literal() -> None:
	y = Var[int, Ctx]("y")
	c10 = Const("Ten", 10)
	assert (c10 >= y).unwrap(ctx) is True
	assert (c10 >= 20).unwrap(ctx) is False


def test_var_in_range() -> None:
	min = Const("min", 0)
	max = Const("max", 10)
	x = Var[int, Ctx]("x")

	expr = (min <= x) & (x <= max)
	assert expr.unwrap(ctx) is True
	assert expr.to_string() == "((min:0 <= x) & (x <= max:10))"
	assert expr.to_string(ctx) == "((min:0 <= x:5 -> True) & (x:5 <= max:10 -> True) -> True)"

	expr = (max >= x) & (x >= min)
	assert expr.unwrap(ctx) is True
	assert expr.to_string() == "((max:10 >= x) & (x >= min:0))"
	assert expr.to_string(ctx) == "((max:10 >= x:5 -> True) & (x:5 >= min:0 -> True) -> True)"

	expr = (min < x) & (x < max)
	assert expr.unwrap(ctx) is True
	assert expr.to_string() == "((min:0 < x) & (x < max:10))"
	assert expr.to_string(ctx) == "((min:0 < x:5 -> True) & (x:5 < max:10 -> True) -> True)"

	expr = (max > x) & (x > min)
	assert expr.unwrap(ctx) is True
	assert expr.to_string() == "((max:10 > x) & (x > min:0))"
	assert expr.to_string(ctx) == "((max:10 > x:5 -> True) & (x:5 > min:0 -> True) -> True)"


def test_deeply_nested_all_operations() -> None:
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	f = Var[float, Ctx]("f")
	c5 = Const("Five", 5)
	c10 = Const("Ten", 10)
	c2 = Const("Two", 2)
	c3 = Const("Three", 3)
	c7 = Const("Seven", 7.0)
	c13 = Const("Thirteen", 13.0)

	expr = ~(
		(((x + c5) * (y - c2)) == c3)
		& ((x >= c5) | (y < c10))
		& ((x != 7) & (y <= 20))
		& ((x > 0) & (y > 0))
		| ((f / c7 + c7) < c13)
	)

	assert_type(expr, Not[Ctx])

	result = expr.unwrap(ctx)
	assert isinstance(result, bool)

	s = expr.to_string()
	print(s)
	assert (
		s
		== "(not (((((((x + Five:5) * (y - Two:2)) == Three:3) & ((x >= Five:5) | (y < Ten:10))) & ((x != 7) & (y <= 20))) & ((x > 0) & (y > 0))) | (((f / Seven:7.0) + Seven:7.0) < Thirteen:13.0)))"
	)

	s_ctx = expr.to_string(ctx)
	print(s_ctx)
	assert (
		s_ctx
		== "(not (((((((x:5 + Five:5 -> 10) * (y:10 - Two:2 -> 8) -> 80) == Three:3 -> False) & ((x:5 >= Five:5 -> True) | (y:10 < Ten:10 -> False) -> True) -> False) & ((x:5 != 7 -> True) & (y:10 <= 20 -> True) -> True) -> False) & ((x:5 > 0 -> True) & (y:10 > 0 -> True) -> True) -> False) | (((f:1.5 / Seven:7.0 -> 0.21428571428571427) + Seven:7.0 -> 7.214285714285714) < Thirteen:13.0 -> True) -> True) -> False)"  # noqa: E501
	)

	assert result is False

	# expression is immutable, so multiple evaluations should yield the same result
	for _ in range(5):
		assert expr.unwrap(ctx) is False
		assert expr.to_string() == s
		assert expr.to_string(ctx) == s_ctx
		assert expr.to_string(ctx) == s_ctx


def test_between() -> None:
	x = Var[int, Ctx]("x")

	expr = between(x, 5, 10)
	assert expr.unwrap(ctx) is False
	assert expr.to_string() == "((Low:5 < x) & (x < High:10))"
	assert expr.to_string(ctx) == "((Low:5 < x:5 -> False) & (x:5 < High:10 -> True) -> False)"

	expr = between(x, 0, 10)
	assert expr.unwrap(ctx) is True
	assert expr.to_string() == "((Low:0 < x) & (x < High:10))"
	assert expr.to_string(ctx) == "((Low:0 < x:5 -> True) & (x:5 < High:10 -> True) -> True)"

	expr = (Const("Low", 5) < x) & (x < Const("High", 10))
	print()
	print(expr.to_string())
	print()
	print(expr.to_string(ctx))


def test_manual_within() -> None:
	f = Var[float, Ctx]("f")

	expr = (f - Const("Target", 1.5)) < Const("Tolerance", 0.01)
	print()
	print(expr.to_string())
	print()
	print(expr.to_string(ctx))


def test_approximately() -> None:
	x = Var[float, Ctx]("x")

	FIVE = PlusMinus("Five", 4.9, 0.1)
	assert_type(FIVE, PlusMinus[float])

	expr = Approximately(x, FIVE)
	assert_type(expr, Approximately[float, Ctx])
	print()
	print(expr.to_string())
	print(expr.to_string(ctx))

	FIVE_ = Percent("Five", 5.0, 1.0)
	assert_type(FIVE_, Percent[float])

	expr = Approximately(x, FIVE_)
	assert_type(expr, Approximately[float, Ctx])

	print()
	print(expr.to_string())
	print(expr.to_string(ctx))


def test_approximately_composition() -> None:
	f = Var[float, Ctx]("f")
	e = Var[float, Ctx]("e")

	f_plus_e = f + e
	SUM = PlusMinus("Sum", 15, 0.1)

	expr = Approximately(f_plus_e, SUM)
	assert_type(expr, Approximately[float, Ctx])

	print()
	print(expr.to_string())
	print(expr.to_string(ctx))

	expr = Approximately(f * e, Percent("Product", 48.0, 5.0))
	assert_type(expr, Approximately[float, Ctx])

	print()
	print(expr.to_string())
	print(expr.to_string(ctx))


def test_composition_nested_arithmetic() -> None:
	"""Test composition of arithmetic expressions at multiple levels."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")

	# Build nested arithmetic expressions
	inner = (x + y) * 2
	outer = inner - 5
	final = outer / 3

	assert final.unwrap(ctx) == ((5 + 10) * 2 - 5) / 3
	assert final.to_string() == "((((x + y) * 2) - 5) / 3)"

	# Test with constants mixed in
	c2 = Const("Two", 2)
	c5 = Const("Five", 5)
	composed = (x + c2) * (y - c5)
	assert composed.unwrap(ctx) == (5 + 2) * (10 - 5)
	assert composed.to_string() == "((x + Two:2) * (y - Five:5))"


def test_composition_comparison_chains() -> None:
	"""Test composition of comparison expressions."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")

	# Chain comparisons with logical operators
	range_check = (0 < x) & (x < 10) & (y > x)
	assert range_check.unwrap(ctx) is True

	complex_comparison = ((x + y) > 10) & ((x * y) < 100)
	assert complex_comparison.unwrap(ctx) is True
	assert complex_comparison.to_string() == "(((x + y) > 10) & ((x * y) < 100))"


def test_composition_mixed_operations() -> None:
	"""Test composition mixing arithmetic, comparison, & logical operations."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")

	# Arithmetic inside comparisons inside logical operations
	expr = ((x + 3) == (y - 2)) | ((x * 2) > y)
	assert expr.unwrap(ctx) is True  # (5+3) == (10-2) is True

	# Nested logical with arithmetic
	complex_expr = ~(((x + y) < 20) & ((x - y) > -10))
	assert complex_expr.unwrap(ctx) is False


def test_composition_with_constants() -> None:
	"""Test composition where constants are used throughout the expression tree."""
	x = Var[int, Ctx]("x")
	base = Const("Base", 10)
	multiplier = Const("Mult", 3)
	threshold = Const("Threshold", 25)

	# Build expression using constants at different levels
	scaled = (x + base) * multiplier
	comparison = scaled > threshold

	assert scaled.unwrap(ctx) == (5 + 10) * 3  # 45
	assert comparison.unwrap(ctx) is True  # 45 > 25
	assert comparison.to_string() == "(((x + Base:10) * Mult:3) > Threshold:25)"


def test_composition_deeply_nested_logical() -> None:
	"""Test deeply nested logical expressions."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")

	# Build deeply nested logical expression
	a = x > 0
	b = y > 0
	c = x < 10
	d = y < 20

	nested = (a & b) | ~(c & d)
	assert nested.unwrap(ctx) is True

	# Even deeper nesting
	deep = ((a & b) | (c & d)) & ~((a | b) & (c | d))
	assert isinstance(deep.unwrap(ctx), bool)


def test_composition_with_function_calls() -> None:
	"""Test composition with special functions like between & approximately."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")

	# Use between in larger expressions
	x_in_range = between(x, 0, 10)
	y_in_range = between(y, 5, 15)
	both_in_range = x_in_range & y_in_range

	assert both_in_range.unwrap(ctx) is True

	# Compose with arithmetic
	sum_expr = x + y
	sum_in_range = between(sum_expr, 10, 20)
	assert sum_in_range.unwrap(ctx) is True


def test_composition_with_approximation() -> None:
	"""Test composition with approximation operations."""
	x = Var[float, Ctx]("x")
	y = Var[float, Ctx]("y")

	# Build expressions that use approximation
	product = x * y
	target = PlusMinus("Target", 50.0, 5.0)
	approx_check = Approximately(product, target)

	# Compose with other conditions
	range_check = (x > 0) & (y > 0)
	full_check = range_check & approx_check

	assert full_check.unwrap(ctx) is True
	assert "≈" in full_check.to_string()
	assert (
		full_check.to_string(ctx)
		== "(((x:5 > 0 -> True) & (y:10 > 0 -> True) -> True) & ((x:5 * y:10 -> 50) ≈ Target:50.0 ± 5.0 -> True) -> True)"
	)  # noqa: E501


def test_composition_reuse_subexpressions() -> None:
	"""Test reusing the same subexpression in multiple places."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")

	# Create a subexpression & reuse it
	sum_expr = x + y
	diff_expr = x - y

	# Use both in different contexts
	sum_condition = sum_expr > 10
	diff_condition = diff_expr < 10
	product_expr = sum_expr * diff_expr
	product_condition = product_expr > 0

	combined = sum_condition & diff_condition & product_condition
	assert combined.unwrap(ctx) is False

	# Verify the subexpressions maintain their identity
	assert sum_expr.unwrap(ctx) == 15
	assert diff_expr.unwrap(ctx) == -5
	assert product_expr.unwrap(ctx) == -75


def test_composition_with_negation() -> None:
	"""Test composition with negation at different levels."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")

	# Negation of simple comparison
	not_equal = ~(x == y)
	assert not_equal.unwrap(ctx) is True

	# Negation of complex expression
	complex_expr = (x > 0) & (y > 0) & (x < y)
	negated_complex = ~complex_expr
	assert negated_complex.unwrap(ctx) is False

	# Double negation
	double_neg = ~~(x == 5)
	assert double_neg.unwrap(ctx) is True


def test_composition_type_consistency() -> None:
	"""Test that type consistency is maintained through composition."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	f = Var[float, Ctx]("f")

	# Mix int & float operations
	mixed_expr = (x + f) > (y * 2.0)
	assert isinstance(mixed_expr.unwrap(ctx), bool)

	# Ensure arithmetic results maintain proper types
	int_result = x + y
	float_result = f * 2.0

	assert isinstance(int_result.unwrap(ctx), int)
	assert isinstance(float_result.unwrap(ctx), float)


def test_composition_immutability() -> None:
	"""Test that composed expressions are immutable."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")

	# Create base expressions
	base_expr = x + y

	# Create composed expressions
	expr1 = base_expr > 10
	expr2 = base_expr < 20

	# Verify that using base_expr in multiple places doesn't affect results
	result1 = expr1.unwrap(ctx)
	result2 = expr2.unwrap(ctx)

	# Re-evaluate to ensure immutability
	assert expr1.unwrap(ctx) == result1
	assert expr2.unwrap(ctx) == result2
	assert base_expr.unwrap(ctx) == 15  # Original value unchanged


def test_composition_string_representation() -> None:
	"""Test that string representations work correctly for composed expressions."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")

	# Build a complex composed expression
	expr = ((x + y) * 2) > ((x - y) + 10)

	# Test symbolic representation
	symbolic = expr.to_string()
	assert "((x + y) * 2)" in symbolic
	assert "((x - y) + 10)" in symbolic
	assert ">" in symbolic

	# Test evaluated representation
	evaluated = expr.to_string(ctx)
	assert "15" in evaluated  # x + y = 15
	assert "30" in evaluated  # (x + y) * 2 = 30
	assert "-5" in evaluated  # x - y = -5
	assert "5" in evaluated  # (x - y) + 10 = 5
	assert "True" in evaluated or "False" in evaluated

	# Repeat using sub expressions
	sum_expr = x + y
	diff_expr = x - y
	sum_expr_x2 = sum_expr * 2
	diff_expr_plus_10 = diff_expr + 10
	expr2 = sum_expr_x2 > diff_expr_plus_10
	assert expr2.to_string() == expr.to_string()
	assert expr2.to_string(ctx) == expr.to_string(ctx)


def test_predicate() -> None:
	"""Test Predicate class."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")

	# Create a predicate
	pred = Predicate("x is greater than y", x > y)

	assert pred.unwrap(ctx) is False  # 5 is not greater than 10
	assert pred.to_string() == "x is greater than y: (x > y)"
	assert pred.to_string(ctx) == "x is greater than y: False (x:5 > y:10 -> False)"

	pred2 = Predicate("x is less than y", x < y)
	assert pred2.unwrap(ctx) is True  # 5 is less than 10
	assert pred2.to_string() == "x is less than y: (x < y)"
	assert pred2.to_string(ctx) == "x is less than y: True (x:5 < y:10 -> True)"

	class Measurement(NamedTuple):
		voltage: float

	voltage = Var[float, Measurement]("voltage")

	voltage_pred = Predicate(
		"Voltage is within range", Approximately(voltage, PlusMinus("Target", 5.0, 0.1))
	)

	assert voltage_pred.unwrap(Measurement(voltage=5.05)) is True
	assert voltage_pred.to_string() == "Voltage is within range: (voltage ≈ Target:5.0 ± 0.1)"
	assert (
		voltage_pred.to_string(Measurement(voltage=5.05))
		== "Voltage is within range: True (voltage:5.05 ≈ Target:5.0 ± 0.1 -> True)"
	)
	assert voltage_pred.unwrap(Measurement(voltage=5.15)) is False
	assert (
		voltage_pred.to_string(Measurement(voltage=5.15))
		== "Voltage is within range: False (voltage:5.15 ≈ Target:5.0 ± 0.1 -> False)"
	)


@pytest.mark.mypy_testing
def test_bound_expr_type() -> None:
	"""Test Closure type with Predicate."""

	class Measurement(NamedTuple):
		voltage: float

	voltage = Var[float, Measurement]("voltage")

	m = Measurement(voltage=5.05)
	voltage_pred = Predicate(
		"Voltage is within range", Approximately(voltage, PlusMinus("Target", 5.0, 0.1))
	)

	closure = voltage_pred.bind(m)
	assert_type(closure, BoundExpr[bool, Measurement])


def test_bind() -> None:
	class Measurement(NamedTuple):
		voltage: float

	voltage = Var[float, Measurement]("voltage")

	m = Measurement(voltage=5.05)

	voltage_pred = Predicate(
		"Voltage is within range", Approximately(voltage, PlusMinus("Target", 5.0, 0.1))
	)

	closure = voltage_pred.bind(m)
	assert closure.unwrap() is True
	assert closure.expr is voltage_pred
	assert closure.ctx is m
	print(closure)
	assert str(closure) == "Voltage is within range: True (voltage:5.05 ≈ Target:5.0 ± 0.1 -> True)"


def test_bind_predicate() -> None:
	"""Test binding a Predicate to a context."""

	class Measurement(NamedTuple):
		voltage: float

	voltage = Var[float, Measurement]("voltage")

	m = Measurement(voltage=5.05)

	expr = Predicate(
		"Voltage is within range", Approximately(voltage, PlusMinus("Target", 5.0, 0.1))
	)
	predicate = expr.bind(m)

	assert predicate.unwrap() is True
	assert predicate.expr is expr
	assert predicate.ctx is m
	print(predicate.ctx)
	print(predicate)
	assert (
		str(predicate) == "Voltage is within range: True (voltage:5.05 ≈ Target:5.0 ± 0.1 -> True)"
	)


def test_pow() -> None:
	"""Test power operation."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")

	# Create a power expression
	pow_expr = x**2
	assert_type(pow_expr, Pow[int, Ctx])

	assert pow_expr.unwrap(ctx) == 25
	assert pow_expr.to_string() == "(x^2)"
	assert pow_expr.to_string(ctx) == "(x:5^2 -> 25)"

	pow_expr = x**y
	assert pow_expr.unwrap(ctx) == 5**10
	assert pow_expr.to_string() == "(x^y)"
	assert pow_expr.to_string(ctx) == "(x:5^y:10 -> 9765625)"
	assert pow_expr.unwrap(ctx) == 9765625

	pow_expr1 = Const(None, 2) ** x
	assert_type(pow_expr1, Pow[int, Any])
	assert pow_expr1.unwrap(ctx) == 2**5
	assert pow_expr1.to_string() == "(2^x)"


def test_approximately_coercion() -> None:
	"""Test coercion of Approximately to BoundExpr."""
	x = Var[float, Ctx]("x")
	target = PlusMinus("Target", 5.0, 0.1)

	expr = target == x
	assert_type(expr, Approximately[float, Ctx])
	print(expr.to_string())

	expr = x == target
	assert_type(expr, Approximately[float, Ctx])
	print(expr.to_string())

	expr1 = target == 5.0
	assert_type(expr1, Approximately[float, Any])
	print(expr1.to_string())



# Copyright (c) 2025 JP Hutchins
# SPDX-License-Identifier: MIT

"""LaTeX serialization for mahonia expressions.

This module provides functionality to convert mahonia expressions to LaTeX
mathematical notation.

>>> from typing import NamedTuple
>>> class Ctx(NamedTuple):
...     x: int
...     y: int
>>> from mahonia import Var, Const
>>> x = Var[int, Ctx]("x")
>>> y = Var[int, Ctx]("y")
>>> expr = x + y * 2
>>> latex(expr)
'x + y \\\\cdot 2'
>>> latex(x > 5)
'x > 5'
"""

from enum import Flag, auto
from typing import Any, Final, Generic, NamedTuple, assert_never

from mahonia import (
	Add,
	And,
	Approximately,
	Const,
	Div,
	Eq,
	Expr,
	Ge,
	Gt,
	Le,
	Lt,
	Mul,
	Ne,
	Not,
	Or,
	Percent,
	PlusMinus,
	Pow,
	Predicate,
	S,
	Sub,
	Var,
)
from mahonia.stats import Count, Mean, Median, Percentile, Range, StdDev

type BinaryOpExpr = (
	Eq[Any, Any]
	| Ne[Any, Any]
	| Lt[Any, Any]
	| Le[Any, Any]
	| Gt[Any, Any]
	| Ge[Any, Any]
	| And[Any, Any]
	| Or[Any, Any]
)


class Show(Flag):
	"""Display options for latex evaluation.

	Assuming that x is 2 and y is 3:
	- (none): `x + y`
	- VALUES: `x:2 + y:3`
	- WORK: `(x + y \\rightarrow 5)`
	- VALUES | WORK: `(x:2 + y:3 \\rightarrow 5)`

	Examples:
	>>> from typing import NamedTuple
	>>> class TestCtx(NamedTuple):
	...     x: int
	...     y: int
	>>> x = Var[int, TestCtx]("x")
	>>> y = Var[int, TestCtx]("y")
	>>> test_ctx = TestCtx(x=2, y=3)
	>>> expr = x + y
	>>> latex(expr, LatexCtx(test_ctx, Show.VALUES))
	'(x:2 + y:3 \\\\rightarrow 5)'
	>>> latex(expr, LatexCtx(test_ctx, Show.WORK))
	'(x + y \\\\rightarrow 5)'
	>>> latex(expr, LatexCtx(test_ctx, Show.VALUES | Show.WORK))
	'(x:2 + y:3 \\\\rightarrow 5)'
	"""

	VALUES = auto()
	"""Add values to variables: `name:<val>."""
	WORK = auto()
	"""Show the evaluated result of the expression."""


class LatexCtx(NamedTuple, Generic[S]):
	ctx: S
	show: Show = Show.VALUES | Show.WORK


def latex(expr: Expr[Any, S], ctx: LatexCtx[S] | None = None) -> str:
	"""Convert a mahonia expression to LaTeX mathematical notation.

	Examples:
	>>> from typing import NamedTuple
	>>> class TestCtx(NamedTuple):
	...     x: float
	...     y: float
	>>> x = Var[float, TestCtx]("x")
	>>> y = Var[float, TestCtx]("y")
	>>> latex(x + y)
	'x + y'
	>>> latex(x * y)
	'x \\\\cdot y'
	>>> latex(x / y)
	'\\\\frac{x}{y}'
	>>> latex(x**2 + y**2)
	'x^2 + y^2'
	>>> # With context - default shows values and work
	>>> test_ctx = TestCtx(x=2.0, y=3.0)
	>>> latex(x + y, LatexCtx(test_ctx))
	'(x:2.0 + y:3.0 \\\\rightarrow 5.0)'
	>>> # Show only values
	>>> latex(x + y, LatexCtx(test_ctx, Show.VALUES))
	'(x:2.0 + y:3.0 \\\\rightarrow 5.0)'
	>>> # Show only work
	>>> latex(x + y, LatexCtx(test_ctx, Show.WORK))
	'(x + y \\\\rightarrow 5.0)'
	>>> # Show nothing (structure only)
	>>> latex(x + y, LatexCtx(test_ctx, Show(0)))
	'(x + y \\\\rightarrow 5.0)'
	>>> # Boolean expressions
	>>> latex(x > y, LatexCtx(test_ctx, Show.VALUES | Show.WORK))
	'(x:2.0 > y:3.0 \\\\rightarrow \\\\text{False})'
	>>> # Complex expressions
	>>> latex((x + y) / 2, LatexCtx(test_ctx, Show.VALUES))
	'(\\\\frac{x:2.0 + y:3.0}{2} \\\\rightarrow 2.5)'
	"""
	match ctx:
		case None:
			return _latex_expr_structure(expr)
		case LatexCtx(ctx_data, show) if show & Show.WORK:
			structure = _latex_expr_structure(expr, ctx)
			return (
				structure
				if "\\rightarrow" in structure
				else _format_with_result(structure, expr.eval(ctx_data))
			)
		case LatexCtx(ctx_data, _):
			return _format_with_result(_latex_expr_structure(expr, ctx), expr.eval(ctx_data))
		case _:
			assert_never(ctx)


def _latex_value(value: Any) -> str:
	"""Convert a value to LaTeX format."""
	if isinstance(value, bool):
		return f"\\text{{{str(value)}}}"
	return str(value)


def _format_with_result(structure: str, result: Const[Any]) -> str:
	"""Format structure with result arrow notation."""
	result_latex = (
		_latex_expr_structure(result)
		if isinstance(result, (PlusMinus, Percent))
		else _latex_value(result.value)
	)
	return f"({structure} \\rightarrow {result_latex})"


def _latex_expr_structure(expr: Expr[Any, Any], latex_ctx: LatexCtx[Any] | None = None) -> str:
	"""Convert expression structure to LaTeX, optionally showing variable values."""
	match expr:
		case Var(name=name):
			match latex_ctx:
				case LatexCtx(ctx, show) if show & Show.VALUES:
					evaluated = expr.eval(ctx)
					return f"{_latex_var(name)}:{evaluated.value}"
				case _:
					return _latex_var(name)
		case PlusMinus(name=name, value=value, plus_minus=pm):
			return f"{_latex_var(name) if name else str(value)} \\pm {pm}"
		case Percent(name=name, value=value, percent=pct):
			return f"{_latex_var(name) if name else str(value)} \\pm {pct}\\%"
		case Const(name=name, value=value) if name:
			match latex_ctx:
				case LatexCtx(_, show) if show & Show.VALUES:
					return f"{_latex_var(name)}:{value}"
				case _:
					return _latex_var(name)
		case Const(value=value):
			return _latex_value(value)
		case Add():
			left = _latex_expr_structure(expr.left, latex_ctx)
			right = _latex_expr_structure(expr.right, latex_ctx)
			match latex_ctx:
				case LatexCtx(ctx, show) if show & Show.WORK:
					return f"({left} + {right} \\rightarrow {_latex_value(expr.eval(ctx).value)})"
				case _:
					return f"{left} + {right}"
		case Sub():
			left = _latex_expr_structure(expr.left, latex_ctx)
			right = _latex_expr_structure(expr.right, latex_ctx)
			right_formatted = f"({right})" if _needs_parentheses(expr.right, expr) else right
			match latex_ctx:
				case LatexCtx(ctx, show) if show & Show.WORK:
					return f"({left} - {right_formatted} \\rightarrow {_latex_value(expr.eval(ctx).value)})"
				case _:
					return f"{left} - {right_formatted}"
		case Mul():
			left = _latex_expr_structure(expr.left, latex_ctx)
			right = _latex_expr_structure(expr.right, latex_ctx)
			left_formatted = f"({left})" if _needs_parentheses(expr.left, expr) else left
			right_formatted = f"({right})" if _needs_parentheses(expr.right, expr) else right
			match latex_ctx:
				case LatexCtx(ctx, show) if show & Show.WORK:
					return f"({left_formatted} \\cdot {right_formatted} \\rightarrow {_latex_value(expr.eval(ctx).value)})"
				case _:
					return f"{left_formatted} \\cdot {right_formatted}"
		case Div():
			left = _latex_expr_structure(expr.left, latex_ctx)
			right = _latex_expr_structure(expr.right, latex_ctx)
			match latex_ctx:
				case LatexCtx(ctx, show) if show & Show.WORK:
					return f"(\\frac{{{left}}}{{{right}}} \\rightarrow {_latex_value(expr.eval(ctx).value)})"
				case _:
					return f"\\frac{{{left}}}{{{right}}}"
		case Pow():
			base = _latex_expr_structure(expr.left, latex_ctx)
			power = _latex_expr_structure(expr.right, latex_ctx)
			formatted_base = f"({base})" if _needs_parentheses(expr.left, expr) else base
			power_formatted = (
				f"{formatted_base}^{{{power}}}"
				if len(power) > 1 or not power.isdigit()
				else f"{formatted_base}^{power}"
			)
			match latex_ctx:
				case LatexCtx(ctx, show) if show & Show.WORK:
					return f"({power_formatted} \\rightarrow {_latex_value(expr.eval(ctx).value)})"
				case _:
					return power_formatted
		case Eq() | Ne() | Lt() | Le() | Gt() | Ge() | And() | Or() as binary_op:
			left = _latex_expr_structure(binary_op.left, latex_ctx)
			right = _latex_expr_structure(binary_op.right, latex_ctx)
			expr_str = f"{left} {LATEX_OP[type(binary_op)]} {right}"
			match latex_ctx:
				case LatexCtx(ctx, show) if show & Show.WORK:
					return f"({expr_str} \\rightarrow {_latex_value(binary_op.eval(ctx).value)})"
				case _:
					return expr_str
		case Not():
			operand = _latex_expr_structure(expr.left, latex_ctx)
			operand_formatted = f"({operand})" if _needs_parentheses(expr.left, expr) else operand
			match latex_ctx:
				case LatexCtx(ctx, show) if show & Show.WORK:
					return f"(\\neg {operand_formatted} \\rightarrow {_latex_value(expr.eval(ctx).value)})"
				case _:
					return f"\\neg {operand_formatted}"
		case Approximately():
			left = _latex_expr_structure(expr.left, latex_ctx)
			right = _latex_expr_structure(expr.right, latex_ctx)
			expr_str = f"{left} \\approx {right}"
			match latex_ctx:
				case LatexCtx(ctx, show) if show & Show.WORK:
					return f"({expr_str} \\rightarrow {_latex_value(expr.eval(ctx).value)})"
				case _:
					return expr_str
		case Predicate(name=name, expr=pred_expr):
			expr_latex = _latex_expr_structure(pred_expr, latex_ctx)
			pred_str = f"\\text{{{name}}}: {expr_latex}" if name else expr_latex
			match latex_ctx:
				case LatexCtx(ctx, show) if show & Show.WORK:
					return f"({pred_str} \\rightarrow {_latex_value(expr.eval(ctx).value)})"
				case _:
					return pred_str
		case Mean():
			operand_formatted = _format_statistical_operand(expr, latex_ctx)
			match latex_ctx:
				case LatexCtx(ctx, show) if show & Show.WORK:
					return f"(\\bar{{{operand_formatted}}} \\rightarrow {_latex_value(expr.eval(ctx).value)})"
				case _:
					return f"\\bar{{{operand_formatted}}}"
		case StdDev():
			operand_formatted = _format_statistical_operand(expr, latex_ctx)
			match latex_ctx:
				case LatexCtx(ctx, show) if show & Show.WORK:
					return f"(\\sigma_{{{operand_formatted}}} \\rightarrow {_latex_value(expr.eval(ctx).value)})"
				case _:
					return f"\\sigma_{{{operand_formatted}}}"
		case Median():
			operand_formatted = _format_statistical_operand(expr, latex_ctx)
			match latex_ctx:
				case LatexCtx(ctx, show) if show & Show.WORK:
					return f"(\\text{{median}}({operand_formatted}) \\rightarrow {_latex_value(expr.eval(ctx).value)})"
				case _:
					return f"\\text{{median}}({operand_formatted})"
		case Percentile():
			operand_formatted = _format_statistical_operand(expr, latex_ctx)
			percentile_str = (
				f"P_{{{int(expr.percentile) if expr.percentile.is_integer() else expr.percentile}}}"
			)
			match latex_ctx:
				case LatexCtx(ctx, show) if show & Show.WORK:
					return f"({percentile_str}({operand_formatted}) \\rightarrow {_latex_value(expr.eval(ctx).value)})"
				case _:
					return f"{percentile_str}({operand_formatted})"
		case Range():
			operand_formatted = _format_statistical_operand(expr, latex_ctx)
			match latex_ctx:
				case LatexCtx(ctx, show) if show & Show.WORK:
					return f"(\\text{{range}}({operand_formatted}) \\rightarrow {_latex_value(expr.eval(ctx).value)})"
				case _:
					return f"\\text{{range}}({operand_formatted})"
		case Count():
			operand_formatted = _format_statistical_operand(expr, latex_ctx)
			match latex_ctx:
				case LatexCtx(ctx, show) if show & Show.WORK:
					return (
						f"(|{operand_formatted}| \\rightarrow {_latex_value(expr.eval(ctx).value)})"
					)
				case _:
					return f"|{operand_formatted}|"
		case _:
			return f"\\text{{Unknown: {type(expr).__name__}}}"


LATEX_OP: Final[dict[type[BinaryOpExpr], str]] = {
	Eq: "=",
	Ne: "\\neq",
	Lt: "<",
	Le: "\\leq",
	Gt: ">",
	Ge: "\\geq",
	And: "\\land",
	Or: "\\lor",
}


def _latex_var(name: str) -> str:
	"""Convert a variable name to LaTeX format.

	Handles subscripts and Greek letters.
	"""
	# Handle subscripts (e.g., x_1 -> x_1, x_max -> x_{max})
	if "_" in name:
		parts = name.split("_", 1)
		base, subscript = parts
		return f"{base}_{subscript}" if len(subscript) == 1 else f"{base}_{{{subscript}}}"

	# Common Greek letters
	return {
		"alpha": "\\alpha",
		"beta": "\\beta",
		"gamma": "\\gamma",
		"delta": "\\delta",
		"epsilon": "\\epsilon",
		"zeta": "\\zeta",
		"eta": "\\eta",
		"theta": "\\theta",
		"iota": "\\iota",
		"kappa": "\\kappa",
		"lambda": "\\lambda",
		"mu": "\\mu",
		"nu": "\\nu",
		"xi": "\\xi",
		"pi": "\\pi",
		"rho": "\\rho",
		"sigma": "\\sigma",
		"tau": "\\tau",
		"upsilon": "\\upsilon",
		"phi": "\\phi",
		"chi": "\\chi",
		"psi": "\\psi",
		"omega": "\\omega",
	}.get(name.lower(), name)


PRECEDENCE: Final[dict[type[Expr[Any, Any]], int]] = {
	Or: 1,
	And: 2,
	Not: 3,
	Eq: 4,
	Ne: 4,
	Lt: 4,
	Le: 4,
	Gt: 4,
	Ge: 4,
	Approximately: 4,
	Add: 5,
	Sub: 5,
	Mul: 6,
	Div: 6,
	Pow: 7,
	Mean: 8,
	StdDev: 8,
	Median: 8,
	Percentile: 8,
	Range: 8,
	Count: 8,
}


def _needs_parentheses(operand: Expr[Any, Any], parent: Expr[Any, Any]) -> bool:
	"""Determine if an operand needs parentheses in the context of its parent operation."""

	if isinstance(
		operand, (Var, Const, PlusMinus, Percent, Mean, StdDev, Median, Percentile, Range, Count)
	):
		return False

	if PRECEDENCE[type(operand)] < PRECEDENCE[type(parent)]:
		return True

	# Special cases for subtraction and division (right-associative concerns)
	if isinstance(parent, Sub) and operand is parent.right:
		if isinstance(operand, (Add, Sub)):
			return True

	if isinstance(parent, Div) and operand is parent.right:
		if isinstance(operand, (Mul, Div)):
			return True

	return False


def _format_statistical_operand(stat_expr: Expr[Any, Any], latex_ctx: LatexCtx[Any] | None) -> str:
	"""Format statistical operand for LaTeX, handling iterable display.

	This leverages the existing to_string method from the statistical operations
	to get proper iterable formatting, then converts to LaTeX-friendly format.
	"""
	match latex_ctx:
		case LatexCtx(ctx, show) if show & Show.VALUES:
			# Use the existing to_string formatting which handles iterable display nicely
			string_repr = stat_expr.to_string(ctx)

			# Extract the operand part from the string representation
			# E.g., "mean(measurements:5[1.0,..5.0] -> 3.0)" -> "measurements:5[1.0,..5.0]"
			if " -> " in string_repr:
				operand_part = string_repr.split(" -> ")[0]
			else:
				operand_part = string_repr

			# Remove the operation name and parentheses
			# E.g., "mean(measurements:5[1.0,..5.0])" -> "measurements:5[1.0,..5.0]"
			# For Percentile: "percentile:95(measurements:5[1.0,..5.0])" -> "measurements:5[1.0,..5.0]"

			# Find the first opening parenthesis and extract what's inside
			paren_start = operand_part.find("(")
			if paren_start != -1:
				# Find the matching closing parenthesis
				paren_end = operand_part.rfind(")")
				if paren_end != -1 and paren_end > paren_start:
					variable_part = operand_part[paren_start + 1 : paren_end]
				else:
					variable_part = operand_part[paren_start + 1 :]
			else:
				variable_part = operand_part

			# Convert variable names to LaTeX format
			# E.g., "measurements:5[1.0,..5.0]" -> "measurements:5[1.0,..5.0]"
			if ":" in variable_part:
				var_name, rest = variable_part.split(":", 1)
				latex_var_name = _latex_var(var_name)
				return f"{latex_var_name}:{rest}"
			else:
				return _latex_var(variable_part)
		case _:
			# Without context or values, just show the variable name
			if hasattr(stat_expr, "left"):
				left_expr = getattr(stat_expr, "left")
				if hasattr(left_expr, "name"):
					return _latex_var(getattr(left_expr, "name"))
			return "data"



# Copyright (c) 2025 JP Hutchins
# SPDX-License-Identifier: MIT

"""
Test that generates LaTeX examples documentation.

This test generates the latex_examples.md file by executing actual Python code
and capturing the LaTeX output. This ensures the documentation stays in sync
with the implementation.
"""

from pathlib import Path
from typing import Any, NamedTuple, cast

from mahonia import (
	Approximately,
	Const,
	Expr,
	Percent,
	PlusMinus,
	Predicate,
	Var,
)
from mahonia.latex import LatexCtx, Show, latex
from mahonia.stats import Count, Mean, Median, Percentile, Range, SizedIterable, StdDev


def test_generate_latex_examples() -> None:
	"""Generate the LaTeX examples markdown file."""

	# Define context for examples
	class Ctx(NamedTuple):
		x: float
		y: float
		z: float
		w: float
		a: float
		b: float
		c: float
		v: float

	# Define statistical context for statistical operations
	class StatsCtx(NamedTuple):
		measurements: SizedIterable[float] = [1.0, 2.0, 3.0, 4.0, 5.0]
		voltages: SizedIterable[float] = [4.95, 5.05, 4.98, 5.02, 5.0]
		batch_sizes: SizedIterable[int] = [10, 12, 8, 15, 11]

	# Start building the markdown content
	lines = [
		"# LaTeX Examples from Mahonia",
		"",
		"This document demonstrates the LaTeX output generated by the `mahonia.latex` module for various mathematical expressions.",
		"",
		"> [!IMPORTANT]",
		"> This file is auto-generated by `tests/test_latex_examples.py`",
		"> Do not edit this file directly. Run the test to regenerate.",
		"",
		"## Setup",
		"",
		"All examples use the following setup:",
		"",
		"```python",
		"from typing import NamedTuple",
		"from mahonia import Var, Const, PlusMinus, Percent, Approximately, Predicate",
		"from mahonia.latex import latex",
		"",
		"class Ctx(NamedTuple):",
		"    x: float",
		"    y: float",
		"    # ... other fields as needed",
		"```",
		"",
		"## Basic Variables",
		"",
		"### Simple Variables",
	]

	# Basic variables
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	lines.extend(
		[
			f"- Variable `x`: `latex(Var[int, Ctx]('x'))` → ${latex(x)}$",
			f"- Variable `y`: `latex(Var[int, Ctx]('y'))` → ${latex(y)}$",
			"",
		]
	)

	# Subscripts
	lines.extend(
		[
			"### Variables with Subscripts",
			"",
		]
	)

	x1 = Var[int, Ctx]("x_1")
	x_max = Var[int, Ctx]("x_max")
	velocity_initial = Var[float, Ctx]("velocity_initial")

	lines.extend(
		[
			f"- Single character subscript: `latex(Var('x_1'))` → ${latex(x1)}$",
			f"- Multiple character subscript: `latex(Var('x_max'))` → ${latex(x_max)}$",
			f"- Complex subscript: `latex(Var('velocity_initial'))` → ${latex(velocity_initial)}$",
			"",
		]
	)

	# Greek letters
	lines.extend(
		[
			"### Greek Letter Variables",
			"",
		]
	)

	greek_vars = [
		("alpha", "α"),
		("beta", "β"),
		("gamma", "γ"),
		("delta", "δ"),
		("epsilon", "ε"),
		("theta", "θ"),
		("pi", "π"),
		("sigma", "σ"),
		("omega", "ω"),
	]

	for name, _symbol in greek_vars:
		var = Var[float, Ctx](name)
		lines.append(f"- {name.capitalize()}: `latex(Var('{name}'))` → ${latex(var)}$")
	lines.extend(["", ""])

	# Constants
	lines.extend(
		[
			"## Constants",
			"",
			"### Numeric Constants",
			"",
		]
	)

	const_examples = [
		(Const(None, 42), "Const(None, 42)", "Integer"),
		(Const(None, 3.14159), "Const(None, 3.14159)", "Float"),
		(Const(None, -5), "Const(None, -5)", "Negative"),
	]

	for const, code, desc in const_examples:
		lines.append(f"- {desc}: `latex({code})` → ${latex(cast(Expr[Any, Any], const))}$")
	lines.extend(["", ""])

	# Named constants
	lines.extend(
		[
			"### Named Constants",
			"",
		]
	)

	pi_const = Const("Pi", 3.14159)  # This becomes \pi due to Greek letter conversion
	zero_const = Const("Zero", 0)

	named_examples = [
		(pi_const, "Const('Pi', 3.14159)", "Named constant Pi (becomes Greek π)"),
		(zero_const, "Const('Zero', 0)", "Zero constant"),
	]

	for const, code, desc in named_examples:
		lines.append(f"- {desc}: `latex({code})` → ${latex(cast(Expr[Any, Any], const))}$")
	lines.extend(["", ""])

	# Arithmetic Operations
	lines.extend(
		[
			"## Arithmetic Operations",
			"",
			"### Basic Operations",
			"",
		]
	)

	x_float = Var[float, Ctx]("x")
	y_float = Var[float, Ctx]("y")

	lines.extend(
		[
			f"- Addition: `latex(x + y)` → ${latex(x + y)}$",
			f"- Subtraction: `latex(x - y)` → ${latex(x - y)}$",
			f"- Multiplication: `latex(x * y)` → ${latex(x_float * y_float)}$",
			f"- Division: `latex(x / y)` → ${latex(x_float / y_float)}$",
			"",
		]
	)

	# Power Operations
	lines.extend(
		[
			"### Power Operations",
			"",
			f"- Square: `latex(x ** 2)` → ${latex(x**2)}$",
			f"- Variable exponent: `latex(x ** y)` → ${latex(x**y)}$",
			f"- Complex exponent: `latex(x ** (y + 1))` → ${latex(x ** (y + 1))}$",
			f"- Base with parentheses: `latex((x + y) ** 2)` → ${latex((x + y) ** 2)}$",
			"",
		]
	)

	# Complex Arithmetic
	lines.extend(
		[
			"### Complex Arithmetic",
			"",
		]
	)

	z_float = Var[float, Ctx]("z")
	complex_examples = [
		(
			((x_float + y_float) * z_float) / (x_float - y_float),
			"((x + y) * z) / (x - y)",
			"Nested expression",
		),
		(
			x_float**3 + Const("Two", 2) * x_float**2 - x_float + Const("One", 1),
			"x**3 + 2*x**2 - x + 1",
			"Polynomial",
		),
	]

	for expr, code, desc in complex_examples:
		if "polynomial" in desc.lower():
			lines.extend(
				[
					"```python",
					"x = Var[float, Ctx]('x')",
					"two = Const('Two', 2)",
					"one = Const('One', 1)",
					"expr = x**3 + two*x**2 - x + one",
					f"latex(expr)  # '{latex(cast(Expr[Any, Any], expr))}'",
					"```",
				]
			)
		else:
			lines.extend(
				[
					"```python",
					"x = Var[float, Ctx]('x')",
					"y = Var[float, Ctx]('y')",
					"z = Var[float, Ctx]('z')",
					f"expr = {code}",
					f"latex(expr)  # '{latex(cast(Expr[Any, Any], expr))}'",
					"```",
				]
			)
		lines.extend(
			[
				f"- {desc}: ${latex(cast(Expr[Any, Any], expr))}$",
				"",
			]
		)

	# Comparison Operations
	lines.extend(
		[
			"## Comparison Operations",
			"",
			"### Basic Comparisons",
			"",
			f"- Equal `latex(x == y)`: ${latex(x == y)}$",
			f"- Not equal `latex(x != y)`: ${latex(x != y)}$",
			f"- Less than `latex(x < y)`: ${latex(x < y)}$",
			f"- Less than or equal `latex(x <= y)`: ${latex(x <= y)}$",
			f"- Greater than `latex(x > y)`: ${latex(x > y)}$",
			f"- Greater than or equal `latex(x >= y)`: ${latex(x >= y)}$",
			"",
			"### Complex Comparisons",
			"",
			f"- With arithmetic `latex((x + 5) > (y * 2))`: ${latex((x + 5) > (y * 2))}$",
			"",
		]
	)

	# Logical Operations
	lines.extend(
		[
			"## Logical Operations",
			"",
			"### Basic Logic",
			"",
			f"- AND `latex((x > 5) & (y < 10))`: ${latex((x > 5) & (y < 10))}$",
			f"- OR `latex((x > 5) | (y < 10))`: ${latex((x > 5) | (y < 10))}$",
			f"- NOT `latex(~(x > 5))`: ${latex(~(x > 5))}$",
			"",
			"### Complex Logic",
			"",
			f"- Nested NOT: ${latex(~((x > 5) & (y < 10)))}$",
			f"- Mixed operations: ${latex(((x_float + y_float) > 10) & ((x_float * y_float) < 50) | (x_float / y_float > 2))}$",
			"",
		]
	)

	# Special Expressions
	pm1 = PlusMinus("measurement", 5.0, 0.1)
	pm2 = PlusMinus(None, 5.0, 0.1)
	pct1 = Percent("measurement", 5.0, 2.0)
	pct2 = Percent(None, 5.0, 2.0)

	x_approx = Var[float, Ctx]("x")
	target1 = PlusMinus("target", 5.0, 0.1)
	target2 = PlusMinus("target2", 10.0, 0.2)
	approx1 = Approximately(x_approx, target1)
	approx2 = Approximately(x_approx ** Const(None, 2.0), target2)

	pred1 = Predicate("condition", x > 5)
	pred2 = Predicate(None, x > 5)

	lines.extend(
		[
			"## Special Expressions",
			"",
			"### Plus-Minus Expressions",
			"",
			f"- Named measurement: ${latex(pm1)}$",
			f"- Unnamed measurement: ${latex(pm2)}$",
			"",
			"### Percentage Expressions",
			"",
			f"- Named percentage: ${latex(pct1)}$",
			f"- Unnamed percentage: ${latex(pct2)}$",
			"",
			"### Approximately Expressions",
			"",
			f"- Basic approximation: ${latex(approx1)}$",
			f"- Complex approximation: ${latex(approx2)}$",
			"",
			"### Predicate Expressions",
			"",
			f"- Named predicate: ${latex(pred1)}$",
			f"- Unnamed predicate: ${latex(pred2)}$",
			"",
		]
	)

	# Real-World Examples
	a = Var[float, Ctx]("a")
	b = Var[float, Ctx]("b")
	c = Var[float, Ctx]("c")
	discriminant = b**2 - Const("Four", 4) * a * c

	m = Var[float, Ctx]("m")
	v = Var[float, Ctx]("v")
	F = Var[float, Ctx]("F")
	half = Const(None, 0.5)
	kinetic_energy = half * m * v**2
	force_eq = F == m * a

	V = Var[float, Ctx]("V")
	voltage_spec = Approximately(V, PlusMinus(None, 5.0, 0.1))
	resistance_tol = Percent("R", 100.0, 5.0)
	T = Var[float, Ctx]("T")
	temp_range = (Const(None, 20) <= T) & (T <= Const(None, 25))

	lines.extend(
		[
			"## Real-World Examples",
			"",
			"### Quadratic Formula Components",
			"",
			f"- Discriminant `b² - 4ac`: ${latex(discriminant)}$",
			"",
			"### Physics Equations",
			"",
			f"- Kinetic energy: ${latex(kinetic_energy)}$",
			f"- Force equation: ${latex(force_eq)}$",
			"",
			"### Engineering Tolerances",
			"",
			f"- Voltage specification: ${latex(voltage_spec)}$",
			f"- Resistance tolerance: ${latex(resistance_tol)}$",
			f"- Temperature range: ${latex(temp_range)}$",
			"",
		]
	)

	# Parentheses Logic Examples
	z_int = Var[int, Ctx]("z")

	lines.extend(
		[
			"## Parentheses Logic Examples",
			"",
			"### Precedence Demonstrations",
			"",
			f"- Multiplication precedence `latex(x + y * z)`: ${latex(x + y * z_int)}$",
			f"- Addition needs parentheses `latex((x + y) * z)`: ${latex((x + y) * z_int)}$",
			f"- Subtraction on right `latex(x - (y + z))`: ${latex(x - (y + z_int))}$",
			f"- Division complexity `latex(x / (y * z))`: ${latex(x_float / (y_float * z_float))}$",
			"",
		]
	)

	# Edge Cases
	zero = Const(None, 0)
	one = Const(None, 1)
	neg_five = Const(None, -5)

	lines.extend(
		[
			"## Edge Cases",
			"",
			"### Special Values",
			"",
			f"- Zero operations `latex(x + 0)`: ${latex(x + zero)}$",
			f"- Negative constants `latex(x + (-5))`: ${latex(x + neg_five)}$",
			f"- Identity operations `latex(x * 1)`: ${latex(x * one)}$",
			"",
		]
	)

	# Statistical Operations Examples - NEW SECTION
	stats_ctx = StatsCtx()

	lines.extend(
		[
			"## Statistical Operations",
			"",
			"Mahonia supports statistical operations on data collections with LaTeX mathematical notation:",
			"",
			"```python",
			"from mahonia.stats import Mean, StdDev, Median, Percentile, Range, Count",
			"",
			"# Define context with sample data",
			"class StatsCtx(NamedTuple):",
			"    measurements: list[float] = [1.0, 2.0, 3.0, 4.0, 5.0]",
			"    voltages: list[float] = [4.95, 5.05, 4.98, 5.02, 5.0]",
			"",
			"# Variables for statistical data",
			"measurements = Var[list[float], StatsCtx]('measurements')",
			"voltages = Var[list[float], StatsCtx]('voltages')",
			"```",
			"",
		]
	)

	# Statistical operations examples
	measurements = Var[SizedIterable[float], StatsCtx]("measurements")
	voltages = Var[SizedIterable[float], StatsCtx]("voltages")

	# Individual operations
	mean_measurements = Mean(measurements)
	stddev_measurements = StdDev(measurements)
	median_measurements = Median(measurements)
	p95_measurements = Percentile(95, measurements)
	range_measurements = Range(measurements)
	count_measurements = Count(measurements)

	lines.extend(
		[
			"### Basic Statistical Operations",
			"",
			f"- Mean (arithmetic average): `latex(Mean(measurements))` → ${latex(mean_measurements)}$",
			f"- Standard deviation: `latex(StdDev(measurements))` → ${latex(stddev_measurements)}$",
			f"- Median: `latex(Median(measurements))` → ${latex(median_measurements)}$",
			f"- 95th percentile: `latex(Percentile(95, measurements))` → ${latex(p95_measurements)}$",
			f"- Range (max - min): `latex(Range(measurements))` → ${latex(range_measurements)}$",
			f"- Count: `latex(Count(measurements))` → ${latex(count_measurements)}$",
			"",
		]
	)

	lines.extend(
		[
			"### Statistical Operations with Context",
			"",
			"When evaluated with context, statistical operations show data values and results:",
			"",
			f"- Mean with data: `latex(Mean(measurements), StatsCtx())` → ${latex(mean_measurements, LatexCtx(stats_ctx))}$",
			f"- Standard deviation with data: `latex(StdDev(measurements), StatsCtx())` → ${latex(stddev_measurements, LatexCtx(stats_ctx))}$",
			f"- Median with data: `latex(Median(measurements), StatsCtx())` → ${latex(median_measurements, LatexCtx(stats_ctx))}$",
			f"- 95th percentile with data: `latex(Percentile(95, measurements), StatsCtx())` → ${latex(p95_measurements, LatexCtx(stats_ctx))}$",
			f"- Range with data: `latex(Range(measurements), StatsCtx())` → ${latex(range_measurements, LatexCtx(stats_ctx))}$",
			f"- Count with data: `latex(Count(measurements), StatsCtx())` → ${latex(count_measurements, LatexCtx(stats_ctx))}$",
			"",
		]
	)

	# Complex statistical expressions
	mean_voltages = Mean(voltages)
	stddev_voltages = StdDev(voltages)
	coefficient_of_variation = stddev_voltages / mean_voltages
	three = Const(None, 3)
	four_nine = Const(None, 4.9)
	process_capability = (mean_voltages - four_nine) / (three * stddev_voltages)

	lines.extend(
		[
			"### Complex Statistical Expressions",
			"",
			"Statistical operations can be combined in complex mathematical expressions:",
			"",
			f"- Coefficient of variation: `σ/μ` → ${latex(coefficient_of_variation)}$",
			f"- Process capability index: `Cp = (μ - LSL)/(3σ)` → ${latex(process_capability)}$",
			"",
			"With context evaluation:",
			f"- CV with data: ${latex(coefficient_of_variation, LatexCtx(stats_ctx))}$",
			f"- Cp with data: ${latex(process_capability, LatexCtx(stats_ctx))}$",
			"",
		]
	)

	# Statistical operations in comparisons and predicates
	four_ninefive = Const(None, 4.95)
	zero_one = Const(None, 0.1)
	quality_check = mean_voltages > four_ninefive
	batch_acceptable = (mean_voltages > four_nine) & (stddev_voltages < zero_one)
	quality_predicate = Predicate("Quality Check", batch_acceptable)

	lines.extend(
		[
			"### Statistical Operations in Quality Control",
			"",
			"Statistical operations are commonly used in quality control and process monitoring:",
			"",
			f"- Quality threshold: `latex(Mean(voltages) > 4.95)` → ${latex(quality_check)}$",
			f"- Batch acceptance: `latex((Mean(voltages) > 4.9) & (StdDev(voltages) < 0.1))` → ${latex(batch_acceptable)}$",
			f"- Quality predicate: `latex(Predicate('Quality Check', condition))` → ${latex(quality_predicate)}$",
			"",
			"With context evaluation:",
			f"- Quality threshold result: ${latex(quality_check, LatexCtx(stats_ctx))}$",
			f"- Batch acceptance result: ${latex(batch_acceptable, LatexCtx(stats_ctx))}$",
			f"- Quality predicate result: ${latex(quality_predicate, LatexCtx(stats_ctx))}$",
			"",
		]
	)

	# Statistical operations with Greek variables - create a context with Greek names
	class GreekStatsCtx(NamedTuple):
		sigma: SizedIterable[float] = [1.0, 2.0, 3.0, 4.0, 5.0]
		alpha: SizedIterable[float] = [1.0, 2.0, 3.0, 4.0, 5.0]

	greek_stats_ctx = GreekStatsCtx()
	sigma_data = Var[SizedIterable[float], GreekStatsCtx]("sigma")
	alpha_data = Var[SizedIterable[float], GreekStatsCtx]("alpha")
	mean_sigma = Mean(sigma_data)
	stddev_alpha = StdDev(alpha_data)

	lines.extend(
		[
			"### Statistical Operations with Greek Variables",
			"",
			"Variable names are automatically converted to Greek letters in LaTeX output:",
			"",
			f"- Mean of sigma data: `latex(Mean(Var('sigma')))` → ${latex(mean_sigma)}$",
			f"- Standard deviation of alpha: `latex(StdDev(Var('alpha')))` → ${latex(stddev_alpha)}$",
			"",
			"With context (using measurements data):",
			f"- Mean sigma with data: ${latex(mean_sigma, LatexCtx(greek_stats_ctx))}$",
			f"- StdDev alpha with data: ${latex(stddev_alpha, LatexCtx(greek_stats_ctx))}$",
			"",
		]
	)

	# Context Examples - NEW SECTION
	# Define a concrete context for examples
	ctx = Ctx(x=2.0, y=3.0, z=1.5, w=4.0, a=1.0, b=2.0, c=3.0, v=5.0)

	lines.extend(
		[
			"## Expressions with Context",
			"",
			"The `latex()` function supports an optional context parameter that evaluates variables and shows their values:",
			"",
			"```python",
			"# Define context with concrete values",
			"ctx = Ctx(x=2.0, y=3.0, z=1.5, w=4.0)",
			"",
			"# Without context - shows expression structure",
			"latex(expr)  # Shows symbolic form",
			"",
			"# With context - shows values",
			"latex(expr, ctx)  # Shows evaluated values",
			"```",
			"",
		]
	)

	# Context examples for variables
	x_ctx = Var[float, Ctx]("x")
	y_ctx = Var[float, Ctx]("y")
	z_ctx = Var[float, Ctx]("z")

	lines.extend(
		[
			"### Variables with Context",
			"",
			f"- Variable without context: `latex(x)` → ${latex(x_ctx)}$",
			f"- Variable with context: `latex(x, LatexCtx(ctx))` → ${latex(x_ctx, LatexCtx(ctx))}$",
			"",
		]
	)

	# Context examples for arithmetic
	lines.extend(
		[
			"### Arithmetic with Context",
			"",
		]
	)

	add_ctx = x_ctx + y_ctx
	mul_ctx = x_ctx * y_ctx
	div_ctx = x_ctx / y_ctx
	pow_ctx = x_ctx**2

	lines.extend(
		[
			f"- Addition: `latex(x + y)` → ${latex(add_ctx)}$",
			f"- Addition with context: `latex(x + y, ctx)` → ${latex(add_ctx, LatexCtx(ctx))}$",
			"",
			f"- Multiplication: `latex(x * y)` → ${latex(mul_ctx)}$",
			f"- Multiplication with context: `latex(x * y, ctx)` → ${latex(mul_ctx, LatexCtx(ctx))}$",
			"",
			f"- Division: `latex(x / y)` → ${latex(div_ctx)}$",
			f"- Division with context: `latex(x / y, ctx)` → ${latex(div_ctx, LatexCtx(ctx))}$",
			"",
			f"- Power: `latex(x ** 2)` → ${latex(pow_ctx)}$",
			f"- Power with context: `latex(x ** 2, ctx)` → ${latex(pow_ctx, LatexCtx(ctx))}$",
			"",
		]
	)

	# Context examples for comparisons and logic
	lines.extend(
		[
			"### Comparisons with Context",
			"",
		]
	)

	lt_ctx = x_ctx < y_ctx
	eq_ctx = x_ctx == 2.0
	complex_logic = (x_ctx > 1.0) & (y_ctx < 5.0)

	lines.extend(
		[
			f"- Less than: `latex(x < y)` → ${latex(lt_ctx)}$",
			f"- Less than with context: `latex(x < y, ctx)` → ${latex(lt_ctx, LatexCtx(ctx))}$",
			"",
			f"- Equality: `latex(x == 2.0)` → ${latex(eq_ctx)}$",
			f"- Equality with context: `latex(x == 2.0, ctx)` → ${latex(eq_ctx, LatexCtx(ctx))}$",
			"",
			f"- Complex logic: `latex((x > 1.0) & (y < 5.0))` → ${latex(complex_logic)}$",
			f"- Complex logic with context: `latex((x > 1.0) & (y < 5.0), ctx)` → ${latex(complex_logic, LatexCtx(ctx))}$",
			"",
		]
	)

	# Context examples for special expressions
	lines.extend(
		[
			"### Special Expressions with Context",
			"",
		]
	)

	approx_ctx = Approximately(x_ctx, PlusMinus("target", 2.0, 0.1))
	pred_ctx = Predicate("condition", x_ctx > 1.5)

	lines.extend(
		[
			f"- Approximately: `latex(x ≈ target ± 0.1)` → ${latex(approx_ctx)}$",
			f"- Approximately with context: `latex(x ≈ target ± 0.1, ctx)` → ${latex(approx_ctx, LatexCtx(ctx))}$",
			"",
			f"- Predicate: `latex(condition: x > 1.5)` → ${latex(pred_ctx)}$",
			f"- Predicate with context: `latex(condition: x > 1.5, ctx)` → ${latex(pred_ctx, LatexCtx(ctx))}$",
			"",
		]
	)

	# NEW SECTION: Show flags functionality
	lines.extend(
		[
			"### Show Flags - Controlling Display Options",
			"",
			"The `latex()` function supports Show flags that control what information is displayed:",
			"",
			"- `Show.VALUES`: Shows variable values as `x:5`",
			"- `Show.WORK`: Shows the evaluation result as `expr → result`",
			"- `Show.VALUES | Show.WORK` (default): Shows both values and result",
			"- `Show(0)`: Shows only structure (no values or work)",
			"",
		]
	)

	# Examples of show_values
	add_show = x_ctx + y_ctx
	comp_show = x_ctx < y_ctx
	complex_show = (x_ctx + z_ctx) * y_ctx

	lines.extend(
		[
			f"- Structure only: `latex(x + y, LatexCtx(ctx, Show(0)))` → ${latex(add_show, LatexCtx(ctx, Show(0)))}$",
			f"- VALUES only: `latex(x + y, LatexCtx(ctx, Show.VALUES))` → ${latex(add_show, LatexCtx(ctx, Show.VALUES))}$",
			f"- WORK only: `latex(x + y, LatexCtx(ctx, Show.WORK))` → ${latex(add_show, LatexCtx(ctx, Show.WORK))}$",
			f"- Both (default): `latex(x + y, LatexCtx(ctx))` → ${latex(add_show, LatexCtx(ctx))}$",
			"",
			"**Boolean Expressions:**",
			f"- Structure: `latex(x < y, LatexCtx(ctx, Show(0)))` → ${latex(comp_show, LatexCtx(ctx, Show(0)))}$",
			f"- VALUES: `latex(x < y, LatexCtx(ctx, Show.VALUES))` → ${latex(comp_show, LatexCtx(ctx, Show.VALUES))}$",
			f"- WORK: `latex(x < y, LatexCtx(ctx, Show.WORK))` → ${latex(comp_show, LatexCtx(ctx, Show.WORK))}$",
			f"- Both: `latex(x < y, LatexCtx(ctx))` → ${latex(comp_show, LatexCtx(ctx))}$",
			"",
			"**Complex Expressions:**",
			f"- Structure: `latex((x + z) * y, LatexCtx(ctx, Show(0)))` → ${latex(complex_show, LatexCtx(ctx, Show(0)))}$",
			f"- VALUES: `latex((x + z) * y, LatexCtx(ctx, Show.VALUES))` → ${latex(complex_show, LatexCtx(ctx, Show.VALUES))}$",
			f"- Both: `latex((x + z) * y, LatexCtx(ctx))` → ${latex(complex_show, LatexCtx(ctx))}$",
			"",
		]
	)

	# Footer
	lines.extend(
		[
			"---",
			"",
			"*Generated by the mahonia LaTeX module test suite*",
			"",
			"To regenerate this file, run: `python -m pytest tests/test_latex_examples.py::test_generate_latex_examples`",
		]
	)

	# Write the markdown file
	output_path = Path(__file__).parent / "latex_examples.md"
	with open(output_path, "w", encoding="utf-8") as f:
		f.write("\n".join(lines))

	print(f"Generated LaTeX examples at {output_path}")


if __name__ == "__main__":
	test_generate_latex_examples()



# Copyright (c) 2025 JP Hutchins
# SPDX-License-Identifier: MIT

"""Property-based tests for statistical operations using hypothesis.

These tests verify mathematical properties and invariants that should hold
for statistical operations, helping catch edge cases and ensure correctness.
"""

import math
import statistics
from typing import NamedTuple

import pytest
from hypothesis import HealthCheck, assume, given, settings
from hypothesis import strategies as st

from mahonia import Approximately, PlusMinus, Predicate, Var
from mahonia.stats import Count, Mean, Median, Percentile, Range, SizedIterable, StdDev


class StatsTestCtx(NamedTuple):
	measurements: list[float]
	batch_id: str


# Strategy for generating non-empty lists of floats
@st.composite
def non_empty_float_lists(draw):
	"""Generate non-empty lists of finite floats for statistical operations."""
	return draw(
		st.lists(
			st.floats(
				min_value=-1000.0,
				max_value=1000.0,
				allow_nan=False,
				allow_infinity=False,
			),
			min_size=1,
			max_size=50,
		)
	)


# Strategy for generating contexts with measurement data
@st.composite
def stats_contexts(draw):
	"""Generate test contexts with measurement data."""
	measurements = draw(non_empty_float_lists())
	batch_id = draw(st.text(min_size=1, max_size=10))
	return StatsTestCtx(measurements=measurements, batch_id=batch_id)


# Strategy for generating contexts with at least 2 measurements (needed for stddev)
@st.composite
def multi_measurement_contexts(draw):
	"""Generate contexts with at least 2 measurements for operations requiring variance."""
	measurements = draw(
		st.lists(
			st.floats(
				min_value=-1000.0,
				max_value=1000.0,
				allow_nan=False,
				allow_infinity=False,
			),
			min_size=2,
			max_size=50,
		)
	)
	batch_id = draw(st.text(min_size=1, max_size=10))
	return StatsTestCtx(measurements=measurements, batch_id=batch_id)


class TestStatisticalProperties:
	"""Test fundamental statistical properties and invariants."""

	@given(stats_contexts())
	def test_mean_bounds(self, ctx: StatsTestCtx):
		"""Mean should be between min and max values."""
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")
		mean_expr = Mean(measurements)

		result = mean_expr.unwrap(ctx)
		min_val = min(ctx.measurements)
		max_val = max(ctx.measurements)

		assert min_val <= result <= max_val

	@given(stats_contexts())
	def test_median_bounds(self, ctx: StatsTestCtx):
		"""Median should be between min and max values."""
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")
		median_expr = Median(measurements)

		result = median_expr.unwrap(ctx)
		min_val = min(ctx.measurements)
		max_val = max(ctx.measurements)

		assert min_val <= result <= max_val

	@given(multi_measurement_contexts())
	def test_stddev_non_negative(self, ctx: StatsTestCtx):
		"""Standard deviation should always be non-negative."""
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")
		stddev_expr = StdDev(measurements)

		result = stddev_expr.unwrap(ctx)
		assert result >= 0.0

	@given(stats_contexts())
	def test_range_non_negative(self, ctx: StatsTestCtx):
		"""Range should always be non-negative."""
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")
		range_expr = Range(measurements)

		result = range_expr.unwrap(ctx)
		assert result >= 0.0

	@given(stats_contexts())
	def test_count_matches_length(self, ctx: StatsTestCtx):
		"""Count should match the actual length of the iterable."""
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")
		count_expr = Count(measurements)

		result = count_expr.unwrap(ctx)
		assert result == len(ctx.measurements)

	@given(
		st.floats(min_value=0.0, max_value=100.0, allow_nan=False, allow_infinity=False),
		stats_contexts(),
	)
	def test_percentile_bounds(self, percentile: float, ctx: StatsTestCtx):
		"""Percentile should be between min and max values."""
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")
		percentile_expr = Percentile(percentile, measurements)

		result = percentile_expr.unwrap(ctx)
		min_val = min(ctx.measurements)
		max_val = max(ctx.measurements)

		# Handle floating point precision issues
		tolerance = 1e-15 * max(1.0, abs(min_val), abs(max_val))
		assert min_val - tolerance <= result <= max_val + tolerance


class TestStatisticalInvariance:
	"""Test mathematical invariance properties of statistical operations."""

	@given(
		stats_contexts(),
		st.floats(min_value=-100.0, max_value=100.0, allow_nan=False, allow_infinity=False),
	)
	def test_mean_translation_invariance(self, ctx: StatsTestCtx, offset: float):
		"""Mean(X + c) = Mean(X) + c for constant c."""
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")

		# Original mean
		mean_expr = Mean(measurements)
		original_mean = mean_expr.unwrap(ctx)

		# Translated measurements
		translated_measurements = [x + offset for x in ctx.measurements]
		translated_ctx = StatsTestCtx(translated_measurements, ctx.batch_id)
		translated_mean = mean_expr.unwrap(translated_ctx)

		# Should differ by exactly the offset
		assert abs(translated_mean - (original_mean + offset)) < 1e-10

	@given(
		stats_contexts(),
		st.floats(min_value=0.1, max_value=10.0, allow_nan=False, allow_infinity=False),
	)
	def test_mean_scale_invariance(self, ctx: StatsTestCtx, scale: float):
		"""Mean(c * X) = c * Mean(X) for positive constant c."""
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")

		# Original mean
		mean_expr = Mean(measurements)
		original_mean = mean_expr.unwrap(ctx)

		# Scaled measurements
		scaled_measurements = [x * scale for x in ctx.measurements]
		scaled_ctx = StatsTestCtx(scaled_measurements, ctx.batch_id)
		scaled_mean = mean_expr.unwrap(scaled_ctx)

		# Should be scaled by the same factor
		expected = original_mean * scale
		# Use absolute tolerance for values near zero, relative for larger values
		tolerance = max(1e-10, 1e-10 * abs(expected))
		assert abs(scaled_mean - expected) < tolerance

	@given(
		multi_measurement_contexts(),
		st.floats(min_value=0.1, max_value=10.0, allow_nan=False, allow_infinity=False),
	)
	def test_stddev_scale_invariance(self, ctx: StatsTestCtx, scale: float):
		"""StdDev(c * X) = |c| * StdDev(X) for constant c."""
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")

		# Original standard deviation
		stddev_expr = StdDev(measurements)
		original_stddev = stddev_expr.unwrap(ctx)

		# Skip if original stddev is very small to avoid precision issues
		assume(original_stddev > 1e-10)

		# Scaled measurements
		scaled_measurements = [x * scale for x in ctx.measurements]
		scaled_ctx = StatsTestCtx(scaled_measurements, ctx.batch_id)
		scaled_stddev = stddev_expr.unwrap(scaled_ctx)

		# Should be scaled by the absolute value of the scale factor
		expected = original_stddev * abs(scale)
		assert abs(scaled_stddev - expected) < 1e-10 * expected

	@given(
		stats_contexts(),
		st.floats(min_value=-100.0, max_value=100.0, allow_nan=False, allow_infinity=False),
	)
	def test_median_translation_invariance(self, ctx: StatsTestCtx, offset: float):
		"""Median(X + c) = Median(X) + c for constant c."""
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")

		# Original median
		median_expr = Median(measurements)
		original_median = median_expr.unwrap(ctx)

		# Translated measurements
		translated_measurements = [x + offset for x in ctx.measurements]
		translated_ctx = StatsTestCtx(translated_measurements, ctx.batch_id)
		translated_median = median_expr.unwrap(translated_ctx)

		# Should differ by exactly the offset
		assert abs(translated_median - (original_median + offset)) < 1e-10


class TestConstantDataProperties:
	"""Test statistical operations on constant data."""

	@given(
		st.floats(min_value=-100.0, max_value=100.0, allow_nan=False, allow_infinity=False),
		st.integers(min_value=1, max_value=20),
	)
	def test_constant_data_mean_equals_value(self, value: float, size: int):
		"""Mean of constant data should equal the constant value."""
		measurements = [value] * size
		ctx = StatsTestCtx(measurements, "CONST")

		measurements_var = Var[SizedIterable[float], StatsTestCtx]("measurements")
		mean_expr = Mean(measurements_var)

		result = mean_expr.unwrap(ctx)
		assert abs(result - value) < 1e-10

	@given(
		st.floats(min_value=-100.0, max_value=100.0, allow_nan=False, allow_infinity=False),
		st.integers(min_value=1, max_value=20),
	)
	def test_constant_data_median_equals_value(self, value: float, size: int):
		"""Median of constant data should equal the constant value."""
		measurements = [value] * size
		ctx = StatsTestCtx(measurements, "CONST")

		measurements_var = Var[SizedIterable[float], StatsTestCtx]("measurements")
		median_expr = Median(measurements_var)

		result = median_expr.unwrap(ctx)
		assert abs(result - value) < 1e-10

	@given(
		st.floats(min_value=-100.0, max_value=100.0, allow_nan=False, allow_infinity=False),
		st.integers(min_value=2, max_value=20),  # Need at least 2 for stddev
	)
	def test_constant_data_stddev_is_zero(self, value: float, size: int):
		"""Standard deviation of constant data should be zero."""
		measurements = [value] * size
		ctx = StatsTestCtx(measurements, "CONST")

		measurements_var = Var[SizedIterable[float], StatsTestCtx]("measurements")
		stddev_expr = StdDev(measurements_var)

		result = stddev_expr.unwrap(ctx)
		assert abs(result) < 1e-10

	@given(
		st.floats(min_value=-100.0, max_value=100.0, allow_nan=False, allow_infinity=False),
		st.integers(min_value=1, max_value=20),
	)
	def test_constant_data_range_is_zero(self, value: float, size: int):
		"""Range of constant data should be zero."""
		measurements = [value] * size
		ctx = StatsTestCtx(measurements, "CONST")

		measurements_var = Var[SizedIterable[float], StatsTestCtx]("measurements")
		range_expr = Range(measurements_var)

		result = range_expr.unwrap(ctx)
		assert abs(result) < 1e-10


class TestSingleElementProperties:
	"""Test statistical operations on single-element data."""

	@given(st.floats(min_value=-100.0, max_value=100.0, allow_nan=False, allow_infinity=False))
	def test_single_element_stats_equal_value(self, value: float):
		"""For single-element data, mean, median should equal the value."""
		ctx = StatsTestCtx([value], "SINGLE")
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")

		mean_expr = Mean(measurements)
		median_expr = Median(measurements)
		range_expr = Range(measurements)
		count_expr = Count(measurements)

		assert abs(mean_expr.unwrap(ctx) - value) < 1e-10
		assert abs(median_expr.unwrap(ctx) - value) < 1e-10
		assert abs(range_expr.unwrap(ctx)) < 1e-10  # Range should be 0
		assert count_expr.unwrap(ctx) == 1

	@given(st.floats(min_value=-100.0, max_value=100.0, allow_nan=False, allow_infinity=False))
	def test_single_element_percentiles(self, value: float):
		"""For single-element data, any percentile should equal the value."""
		ctx = StatsTestCtx([value], "SINGLE")
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")

		# Test various percentiles
		for percentile in [0, 25, 50, 75, 95, 100]:
			percentile_expr = Percentile(percentile, measurements)
			result = percentile_expr.unwrap(ctx)
			assert abs(result - value) < 1e-10


class TestArithmeticProperties:
	"""Test arithmetic operations with statistical functions."""

	@given(stats_contexts())
	def test_mean_arithmetic_consistency(self, ctx: StatsTestCtx):
		"""Test arithmetic operations with mean expressions."""
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")
		mean_expr = Mean(measurements)

		# Test mean + constant
		const_val = 10.0
		sum_expr = mean_expr + const_val
		result = sum_expr.unwrap(ctx)
		expected = mean_expr.unwrap(ctx) + const_val
		assert abs(result - expected) < 1e-10

		# Test mean * constant
		mult_expr = mean_expr * 2.0
		mult_result = mult_expr.unwrap(ctx)
		expected_mult = mean_expr.unwrap(ctx) * 2.0
		assert abs(mult_result - expected_mult) < 1e-10

	@given(multi_measurement_contexts())
	def test_statistical_operation_composition(self, ctx: StatsTestCtx):
		"""Test composing different statistical operations."""
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")

		mean_expr = Mean(measurements)
		stddev_expr = StdDev(measurements)

		# Coefficient of variation: stddev / mean (if mean != 0)
		mean_val = mean_expr.unwrap(ctx)
		stddev_val = stddev_expr.unwrap(ctx)
		assume(abs(mean_val) > 1e-10)  # Avoid division by near-zero
		assume(abs(stddev_val) > 1e-12 or abs(mean_val) > 1e-6)  # Avoid both being zero

		cv_expr = stddev_expr / mean_expr
		cv_result = cv_expr.unwrap(ctx)

		expected_cv = stddev_val / mean_val
		# Use absolute tolerance for small values, relative for larger values
		tolerance = max(1e-12, 1e-10 * abs(expected_cv))
		assert abs(cv_result - expected_cv) < tolerance


class TestStatisticalComparisons:
	"""Test comparison operations with statistical functions."""

	@given(stats_contexts())
	def test_mean_with_tolerance_checking(self, ctx: StatsTestCtx):
		"""Test using statistical operations with tolerance specifications."""
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")
		mean_expr = Mean(measurements)

		# Create a tolerance around the actual mean
		actual_mean = mean_expr.unwrap(ctx)
		tolerance = 1.0
		spec = PlusMinus("Spec", actual_mean, tolerance)

		# Should be approximately equal
		approx_expr = Approximately(mean_expr, spec)
		assert approx_expr.unwrap(ctx) is True

		# Test with tighter tolerance
		tight_spec = PlusMinus("Tight", actual_mean, 0.001)
		tight_approx = Approximately(mean_expr, tight_spec)
		assert tight_approx.unwrap(ctx) is True

	@given(multi_measurement_contexts())
	def test_statistical_predicate_composition(self, ctx: StatsTestCtx):
		"""Test complex predicates using multiple statistical operations."""
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")

		mean_expr = Mean(measurements)
		stddev_expr = StdDev(measurements)
		count_expr = Count(measurements)

		# Create a predicate: mean > 0 AND stddev < 1000 AND count >= 1
		predicate = Predicate(
			"Quality Check", (mean_expr > -1000.0) & (stddev_expr < 1000.0) & (count_expr >= 1)
		)

		# This should generally be true for our test data
		result = predicate.unwrap(ctx)
		assert isinstance(result, bool)


class TestStringRepresentationProperties:
	"""Test string representation properties of statistical operations."""

	@given(stats_contexts())
	def test_string_contains_operation_name(self, ctx: StatsTestCtx):
		"""String representation should contain the operation name."""
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")

		operations = [
			(Mean(measurements), "mean"),
			(Median(measurements), "median"),
			(Range(measurements), "range"),
			(Count(measurements), "count"),
		]

		# Only test StdDev if we have enough data
		if len(ctx.measurements) >= 2:
			operations.append((StdDev(measurements), "stddev"))

		for expr, op_name in operations:
			str_repr = expr.to_string()
			assert op_name in str_repr.lower()

			# With context, should show evaluation
			eval_str = expr.to_string(ctx)
			assert op_name in eval_str.lower()
			assert "->" in eval_str

	@given(
		stats_contexts(),
		st.floats(min_value=0.0, max_value=100.0, allow_nan=False, allow_infinity=False),
	)
	def test_percentile_string_contains_value(self, ctx: StatsTestCtx, percentile: float):
		"""Percentile string representation should contain the percentile value."""
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")
		percentile_expr = Percentile(percentile, measurements)

		str_repr = percentile_expr.to_string()
		assert "percentile" in str_repr.lower()
		assert str(percentile) in str_repr


class TestEdgeCases:
	"""Test edge cases and boundary conditions."""

	def test_empty_list_handling(self):
		"""Test that empty lists are handled appropriately."""
		empty_ctx = StatsTestCtx([], "EMPTY")
		measurements = Var[SizedIterable[float], StatsTestCtx]("measurements")

		# Count should work with empty data
		count_expr = Count(measurements)
		assert count_expr.unwrap(empty_ctx) == 0

		# Statistical operations should raise appropriate errors
		mean_expr = Mean(measurements)
		with pytest.raises(statistics.StatisticsError):
			mean_expr.unwrap(empty_ctx)

	@given(
		st.lists(
			st.floats(min_value=-1e6, max_value=1e6, allow_nan=False, allow_infinity=False),
			min_size=1,
			max_size=3,
		)
	)
	def test_extreme_values_handling(self, measurements: list[float]):
		"""Test handling of extreme but finite values."""
		ctx = StatsTestCtx(measurements, "EXTREME")
		measurements_var = Var[SizedIterable[float], StatsTestCtx]("measurements")

		# All basic operations should work with extreme values
		mean_expr = Mean(measurements_var)
		median_expr = Median(measurements_var)
		range_expr = Range(measurements_var)
		count_expr = Count(measurements_var)

		# Should not raise exceptions
		mean_result = mean_expr.unwrap(ctx)
		median_result = median_expr.unwrap(ctx)
		range_result = range_expr.unwrap(ctx)
		count_result = count_expr.unwrap(ctx)

		# Results should be finite
		assert math.isfinite(mean_result)
		assert math.isfinite(median_result)
		assert math.isfinite(range_result)
		assert count_result == len(measurements)

	@given(non_empty_float_lists())
	@settings(suppress_health_check=[HealthCheck.filter_too_much])
	def test_ordering_invariance(self, measurements: list[float]):
		"""Test that statistical results are invariant to ordering."""
		# Only test if we have more than one unique value
		unique_vals = list(set(measurements))
		assume(len(unique_vals) > 1)

		original_ctx = StatsTestCtx(measurements, "ORIGINAL")
		shuffled_ctx = StatsTestCtx(measurements[::-1], "SHUFFLED")  # Simple reverse

		measurements_var = Var[SizedIterable[float], StatsTestCtx]("measurements")

		# These operations should be order-invariant
		mean_op = Mean(measurements_var)
		median_op = Median(measurements_var)
		range_op = Range(measurements_var)
		count_op = Count(measurements_var)

		# Test each operation
		mean_original = mean_op.unwrap(original_ctx)
		mean_shuffled = mean_op.unwrap(shuffled_ctx)
		assert abs(mean_original - mean_shuffled) < 1e-10 * max(1.0, abs(mean_original))

		median_original = median_op.unwrap(original_ctx)
		median_shuffled = median_op.unwrap(shuffled_ctx)
		assert abs(median_original - median_shuffled) < 1e-10 * max(1.0, abs(median_original))

		range_original = range_op.unwrap(original_ctx)
		range_shuffled = range_op.unwrap(shuffled_ctx)
		assert abs(range_original - range_shuffled) < 1e-10 * max(1.0, abs(range_original))

		assert count_op.unwrap(original_ctx) == count_op.unwrap(shuffled_ctx)

		# Only test StdDev if we have enough data
		if len(measurements) >= 2:
			stddev_op = StdDev(measurements_var)
			stddev_original = stddev_op.unwrap(original_ctx)
			stddev_shuffled = stddev_op.unwrap(shuffled_ctx)
			assert abs(stddev_original - stddev_shuffled) < 1e-10 * max(1.0, abs(stddev_original))



# Copyright (c) 2025 JP Hutchins
# SPDX-License-Identifier: MIT

from typing import NamedTuple

from mahonia import (
	Approximately,
	Const,
	Percent,
	PlusMinus,
	Predicate,
	Var,
)
from mahonia.latex import LatexCtx, Show, latex
from mahonia.stats import Count, Mean, Median, Percentile, Range, SizedIterable, StdDev


class Ctx(NamedTuple):
	x: int
	y: int
	f: float = 1.5
	name: str = "test"
	alpha: float = 2.5
	beta: float = 3.0
	gamma: float = 4.0
	delta: float = 5.0
	theta: float = 6.0
	pi: float = 3.14159
	sigma: float = 7.0
	omega: float = 8.0


ctx = Ctx(x=5, y=10)


class StatsCtx(NamedTuple):
	values: SizedIterable[float] = [1.0, 2.0, 3.0, 4.0, 5.0]
	measurements: SizedIterable[float] = [10.0, 15.0, 20.0, 25.0, 30.0]


stats_ctx = StatsCtx()


# Basic Expressions Tests


def test_var_simple() -> None:
	"""Test simple variable conversion."""
	x = Var[int, Ctx]("x")
	assert latex(x) == "x"


def test_var_subscript_single() -> None:
	"""Test variable with single character subscript."""
	x1 = Var[int, Ctx]("x_1")
	assert latex(x1) == "x_1"


def test_var_subscript_multiple() -> None:
	"""Test variable with multiple character subscript."""
	x_max = Var[int, Ctx]("x_max")
	assert latex(x_max) == "x_{max}"


def test_var_greek_letters() -> None:
	"""Test Greek letter variable names."""
	alpha = Var[float, Ctx]("alpha")
	beta = Var[float, Ctx]("beta")
	gamma = Var[float, Ctx]("gamma")
	delta = Var[float, Ctx]("delta")
	epsilon = Var[float, Ctx]("epsilon")
	theta = Var[float, Ctx]("theta")
	pi = Var[float, Ctx]("pi")
	sigma = Var[float, Ctx]("sigma")
	omega = Var[float, Ctx]("omega")

	assert latex(alpha) == "\\alpha"
	assert latex(beta) == "\\beta"
	assert latex(gamma) == "\\gamma"
	assert latex(delta) == "\\delta"
	assert latex(epsilon) == "\\epsilon"
	assert latex(theta) == "\\theta"
	assert latex(pi) == "\\pi"
	assert latex(sigma) == "\\sigma"
	assert latex(omega) == "\\omega"


def test_const_with_name() -> None:
	"""Test constant with name."""
	c = Const("Pi", 3.14159)
	assert latex(c) == "\\pi"  # "Pi" gets converted to Greek letter π


def test_const_without_name() -> None:
	"""Test constant without name."""
	c = Const(None, 42)
	assert latex(c) == "42"


def test_const_float() -> None:
	"""Test float constant."""
	c = Const(None, 3.14)
	assert latex(c) == "3.14"


def test_const_negative() -> None:
	"""Test negative constant."""
	c = Const(None, -5)
	assert latex(c) == "-5"


# Arithmetic Operations Tests


def test_addition() -> None:
	"""Test addition operation."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = x + y
	assert latex(expr) == "x + y"


def test_addition_with_constant() -> None:
	"""Test addition with constant."""
	x = Var[int, Ctx]("x")
	c = Const("Five", 5)
	expr = x + c
	assert latex(expr) == "x + Five"


def test_subtraction() -> None:
	"""Test subtraction operation."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = x - y
	assert latex(expr) == "x - y"


def test_subtraction_with_parentheses() -> None:
	"""Test subtraction that requires parentheses."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	z = Var[int, Ctx]("z")
	expr = x - (y + z)
	assert latex(expr) == "x - (y + z)"


def test_multiplication() -> None:
	"""Test multiplication operation."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = x * y
	assert latex(expr) == "x \\cdot y"


def test_multiplication_with_parentheses() -> None:
	"""Test multiplication that requires parentheses."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	z = Var[int, Ctx]("z")
	expr = (x + y) * (z - x)
	assert latex(expr) == "(x + y) \\cdot (z - x)"


def test_division() -> None:
	"""Test division operation."""
	x = Var[float, Ctx]("x")
	y = Var[float, Ctx]("y")
	expr = x / y
	assert latex(expr) == "\\frac{x}{y}"


def test_division_complex() -> None:
	"""Test division with complex expressions."""
	x = Var[float, Ctx]("x")
	y = Var[float, Ctx]("y")
	z = Var[float, Ctx]("z")
	expr = (x + y) / (z - x)
	assert latex(expr) == "\\frac{x + y}{z - x}"


def test_power_simple() -> None:
	"""Test power operation."""
	x = Var[int, Ctx]("x")
	expr = x**2
	assert latex(expr) == "x^2"


def test_power_variable_exponent() -> None:
	"""Test power with variable exponent."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = x**y
	assert latex(expr) == "x^{y}"


def test_power_complex_exponent() -> None:
	"""Test power with complex exponent."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = x ** (y + 1)
	assert latex(expr) == "x^{y + 1}"


def test_power_with_parentheses() -> None:
	"""Test power that requires parentheses for base."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = (x + y) ** 2
	assert latex(expr) == "(x + y)^2"


# Comparison Operations Tests


def test_equality() -> None:
	"""Test equality operator."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = x == y
	assert latex(expr) == "x = y"


def test_inequality() -> None:
	"""Test inequality operator."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = x != y
	assert latex(expr) == "x \\neq y"


def test_less_than() -> None:
	"""Test less than operator."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = x < y
	assert latex(expr) == "x < y"


def test_less_equal() -> None:
	"""Test less than or equal operator."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = x <= y
	assert latex(expr) == "x \\leq y"


def test_greater_than() -> None:
	"""Test greater than operator."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = x > y
	assert latex(expr) == "x > y"


def test_greater_equal() -> None:
	"""Test greater than or equal operator."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = x >= y
	assert latex(expr) == "x \\geq y"


def test_comparison_with_expression() -> None:
	"""Test comparison with arithmetic expressions."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = (x + 5) > (y * 2)
	assert latex(expr) == "x + 5 > y \\cdot 2"


# Logical Operations Tests


def test_and_operation() -> None:
	"""Test logical AND operation."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = (x > 5) & (y < 10)
	assert latex(expr) == "x > 5 \\land y < 10"


def test_or_operation() -> None:
	"""Test logical OR operation."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = (x > 5) | (y < 10)
	assert latex(expr) == "x > 5 \\lor y < 10"


def test_not_operation() -> None:
	"""Test logical NOT operation."""
	x = Var[int, Ctx]("x")
	expr = ~(x > 5)
	assert latex(expr) == "\\neg x > 5"


def test_not_with_parentheses() -> None:
	"""Test NOT operation that requires parentheses."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = ~((x > 5) & (y < 10))
	assert latex(expr) == "\\neg (x > 5 \\land y < 10)"


def test_complex_logical_expression() -> None:
	"""Test complex logical expression."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	z = Var[int, Ctx]("z")
	expr = ((x > 0) & (y > 0)) | (z < 0)
	assert latex(expr) == "x > 0 \\land y > 0 \\lor z < 0"


# Special Expressions Tests


def test_plus_minus() -> None:
	"""Test plus-minus expression."""
	pm = PlusMinus("measurement", 5.0, 0.1)
	assert latex(pm) == "measurement \\pm 0.1"


def test_plus_minus_without_name() -> None:
	"""Test plus-minus without name."""
	pm = PlusMinus(None, 5.0, 0.1)
	assert latex(pm) == "5.0 \\pm 0.1"


def test_percent() -> None:
	"""Test percent expression."""
	pct = Percent("measurement", 5.0, 2.0)
	assert latex(pct) == "measurement \\pm 2.0\\%"


def test_percent_without_name() -> None:
	"""Test percent without name."""
	pct = Percent(None, 5.0, 2.0)
	assert latex(pct) == "5.0 \\pm 2.0\\%"


def test_approximately() -> None:
	"""Test approximately operation."""
	x = Var[float, Ctx]("x")
	target = PlusMinus("target", 5.0, 0.1)
	expr = Approximately(x, target)
	assert latex(expr) == "x \\approx target \\pm 0.1"


def test_predicate_with_name() -> None:
	"""Test predicate with name."""
	x = Var[int, Ctx]("x")
	pred = Predicate("condition", x > 5)
	assert latex(pred) == "\\text{condition}: x > 5"


def test_predicate_without_name() -> None:
	"""Test predicate without name."""
	x = Var[int, Ctx]("x")
	pred = Predicate(None, x > 5)
	assert latex(pred) == "x > 5"


# Complex Expressions Tests


def test_nested_arithmetic() -> None:
	"""Test deeply nested arithmetic expressions."""
	x = Var[float, Ctx]("x")
	y = Var[float, Ctx]("y")
	z = Var[float, Ctx]("z")
	expr = ((x + y) * z) / (x - y)
	assert latex(expr) == "\\frac{(x + y) \\cdot z}{x - y}"


def test_polynomial_expression() -> None:
	"""Test polynomial-like expression."""
	x = Var[float, Ctx]("x")
	two = Const("Two", 2)
	one = Const("One", 1)
	expr = x**3 + two * x**2 - x + one
	assert latex(expr) == "x^3 + Two \\cdot x^2 - x + One"


def test_quadratic_formula() -> None:
	"""Test quadratic formula representation."""
	a = Var[float, Ctx]("a")
	b = Var[float, Ctx]("b")
	c = Var[float, Ctx]("c")
	four = Const("Four", 4)
	two = Const("Two", 2)
	# x = (-b ± sqrt(b² - 4ac)) / (2a)
	discriminant = b**2 - four * a * c
	neg_b = Const("NegB", -1) * b
	expr = (neg_b + discriminant) / (two * a)  # Just the positive part
	assert latex(expr) == "\\frac{NegB \\cdot b + b^2 - Four \\cdot a \\cdot c}{Two \\cdot a}"


def test_complex_logical_with_arithmetic() -> None:
	"""Test complex expression mixing logical and arithmetic operations."""
	x = Var[float, Ctx]("x")
	y = Var[float, Ctx]("y")
	expr = ((x + y) > 10) & ((x * y) < 50) | (x / y > 2)
	assert latex(expr) == "x + y > 10 \\land x \\cdot y < 50 \\lor \\frac{x}{y} > 2"


def test_nested_function_composition() -> None:
	"""Test nested function-like expressions."""
	x = Var[float, Ctx]("x")
	target1 = PlusMinus("target1", 5.0, 0.1)
	target2 = PlusMinus("target2", 10.0, 0.2)
	expr1: Approximately[float, Ctx] = Approximately(x, target1)
	two = Const(None, 2.0)  # Use float constant to maintain float type
	power_expr = x**two
	expr2: Approximately[float, Ctx] = Approximately(power_expr, target2)
	combined = expr1 & expr2
	assert latex(combined) == "x \\approx target1 \\pm 0.1 \\land x^{2.0} \\approx target2 \\pm 0.2"


def test_deeply_nested_parentheses() -> None:
	"""Test expression that requires multiple levels of parentheses."""
	x = Var[float, Ctx]("x")
	y = Var[float, Ctx]("y")
	z = Var[float, Ctx]("z")
	w = Var[float, Ctx]("w")
	expr = (x + (y - (z * w))) / (x - (y + z))
	assert latex(expr) == "\\frac{x + y - z \\cdot w}{x - (y + z)}"


# Parentheses Logic Tests


def test_addition_in_multiplication() -> None:
	"""Test addition inside multiplication needs parentheses."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	z = Var[int, Ctx]("z")
	expr = (x + y) * z
	assert latex(expr) == "(x + y) \\cdot z"


def test_multiplication_in_addition() -> None:
	"""Test multiplication inside addition doesn't need parentheses."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	z = Var[int, Ctx]("z")
	expr = x + y * z
	assert latex(expr) == "x + y \\cdot z"


def test_subtraction_right_side_parentheses() -> None:
	"""Test right side of subtraction needs parentheses for addition/subtraction."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	z = Var[int, Ctx]("z")
	expr1 = x - (y + z)
	expr2 = x - (y - z)
	assert latex(expr1) == "x - (y + z)"
	assert latex(expr2) == "x - (y - z)"


def test_division_right_side_parentheses() -> None:
	"""Test right side of division needs parentheses for multiplication/division."""
	x = Var[float, Ctx]("x")
	y = Var[float, Ctx]("y")
	z = Var[float, Ctx]("z")
	expr1 = x / (y * z)
	expr2 = x / (y / z)
	assert latex(expr1) == "\\frac{x}{y \\cdot z}"
	assert latex(expr2) == "\\frac{x}{\\frac{y}{z}}"


def test_power_base_parentheses() -> None:
	"""Test power base needs parentheses for lower precedence operations."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = (x + y) ** 2
	assert latex(expr) == "(x + y)^2"


def test_logical_precedence() -> None:
	"""Test logical operator precedence."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	z = Var[int, Ctx]("z")
	# OR has lower precedence than AND
	expr = (x > 0) | (y > 0) & (z > 0)
	assert latex(expr) == "x > 0 \\lor y > 0 \\land z > 0"


# Edge Cases Tests


def test_unknown_expression_type() -> None:
	"""Test handling of unknown expression types."""
	# This is tricky to test since we can't easily create unknown types
	# In real usage, this would be handled by the default case
	pass


def test_empty_variable_name() -> None:
	"""Test variable with empty name."""
	x = Var[int, Ctx]("")
	assert latex(x) == ""


def test_special_characters_in_names() -> None:
	"""Test special characters in variable names."""
	x = Var[int, Ctx]("x_prime")
	assert latex(x) == "x_{prime}"


def test_zero_values() -> None:
	"""Test expressions with zero values."""
	zero = Const(None, 0)
	x = Var[int, Ctx]("x")
	expr1 = x + zero
	expr2 = x * zero
	assert latex(expr1) == "x + 0"
	assert latex(expr2) == "x \\cdot 0"


def test_negative_constants_in_expressions() -> None:
	"""Test negative constants in complex expressions."""
	x = Var[int, Ctx]("x")
	neg = Const(None, -5)
	expr = x + neg
	assert latex(expr) == "x + -5"


def test_very_long_variable_names() -> None:
	"""Test very long variable names."""
	long_name = "very_long_variable_name_with_many_characters"
	x = Var[int, Ctx](long_name)
	assert latex(x) == "very_{long_variable_name_with_many_characters}"


# Docstring Examples Tests


def test_docstring_examples() -> None:
	"""Test the examples from the module docstring."""
	from typing import NamedTuple

	class TestCtx(NamedTuple):
		x: int
		y: int

	x = Var[int, TestCtx]("x")
	y = Var[int, TestCtx]("y")

	expr = x + y * 2
	assert latex(expr) == "x + y \\cdot 2"

	expr2 = x > 5
	assert latex(expr2) == "x > 5"


def test_function_docstring_examples() -> None:
	"""Test the examples from the latex function docstring."""
	from typing import NamedTuple

	class TestCtx(NamedTuple):
		x: float
		y: float

	x = Var[float, TestCtx]("x")
	y = Var[float, TestCtx]("y")

	# Test basic addition
	assert latex(x + y) == "x + y"

	# Test multiplication
	assert latex(x * y) == "x \\cdot y"

	# Test division
	assert latex(x / y) == "\\frac{x}{y}"

	# Test power
	assert latex(x**2 + y**2) == "x^2 + y^2"


# Evaluated Expression Tests


def test_latex_with_context_variables() -> None:
	"""Test LaTeX output of variables with context."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")

	# Without context
	assert latex(x) == "x"
	assert latex(y) == "y"

	# With context - shows expression -> result
	assert latex(x, LatexCtx(ctx)) == "(x:5 \\rightarrow 5)"
	assert latex(y, LatexCtx(ctx)) == "(y:10 \\rightarrow 10)"


def test_latex_with_context_arithmetic() -> None:
	"""Test LaTeX output of arithmetic expressions with context."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")

	# Addition
	add_expr = x + y
	assert latex(add_expr) == "x + y"
	assert latex(add_expr, LatexCtx(ctx)) == "(x:5 + y:10 \\rightarrow 15)"

	# Subtraction
	sub_expr = y - x
	assert latex(sub_expr) == "y - x"
	assert latex(sub_expr, LatexCtx(ctx)) == "(y:10 - x:5 \\rightarrow 5)"

	# Multiplication
	mul_expr = x * y
	assert latex(mul_expr) == "x \\cdot y"
	assert latex(mul_expr, LatexCtx(ctx)) == "(x:5 \\cdot y:10 \\rightarrow 50)"

	# Division (need floats for proper division)
	x_float = Var[float, Ctx]("f")  # f = 1.5
	y_float = Var[float, Ctx]("beta")  # beta = 3.0
	div_expr = x_float / y_float
	assert latex(div_expr) == "\\frac{f}{\\beta}"
	assert latex(div_expr, LatexCtx(ctx)) == "(\\frac{f:1.5}{\\beta:3.0} \\rightarrow 0.5)"


def test_latex_with_context_power() -> None:
	"""Test LaTeX output of power expressions with context."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")

	# Simple power
	power_expr = x**2
	assert latex(power_expr) == "x^2"
	assert latex(power_expr, LatexCtx(ctx)) == "(x:5^2 \\rightarrow 25)"

	# Variable exponent
	var_power_expr = x**y
	assert latex(var_power_expr) == "x^{y}"
	assert latex(var_power_expr, LatexCtx(ctx)) == "(x:5^{y:10} \\rightarrow 9765625)"


def test_latex_with_context_comparisons() -> None:
	"""Test LaTeX output of comparison expressions with context."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")

	# Equal
	eq_expr = x == 5
	assert latex(eq_expr) == "x = 5"
	assert latex(eq_expr, LatexCtx(ctx)) == "(x:5 = 5 \\rightarrow \\text{True})"

	# Not equal
	ne_expr = x != y
	assert latex(ne_expr) == "x \\neq y"
	assert latex(ne_expr, LatexCtx(ctx)) == "(x:5 \\neq y:10 \\rightarrow \\text{True})"

	# Less than
	lt_expr = x < y
	assert latex(lt_expr) == "x < y"
	assert latex(lt_expr, LatexCtx(ctx)) == "(x:5 < y:10 \\rightarrow \\text{True})"

	# Greater than
	gt_expr = x > y
	assert latex(gt_expr) == "x > y"
	assert latex(gt_expr, LatexCtx(ctx)) == "(x:5 > y:10 \\rightarrow \\text{False})"


def test_latex_with_context_logical() -> None:
	"""Test LaTeX output of logical expressions with context."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")

	# AND
	and_expr = (x > 0) & (y > 0)
	assert latex(and_expr) == "x > 0 \\land y > 0"
	assert (
		latex(and_expr, LatexCtx(ctx))
		== "((x:5 > 0 \\rightarrow \\text{True}) \\land (y:10 > 0 \\rightarrow \\text{True}) \\rightarrow \\text{True})"
	)

	# OR
	or_expr = (x > 100) | (y > 0)
	assert latex(or_expr) == "x > 100 \\lor y > 0"
	assert (
		latex(or_expr, LatexCtx(ctx))
		== "((x:5 > 100 \\rightarrow \\text{False}) \\lor (y:10 > 0 \\rightarrow \\text{True}) \\rightarrow \\text{True})"
	)

	# NOT
	not_expr = ~(x > 100)
	assert latex(not_expr) == "\\neg x > 100"
	assert (
		latex(not_expr, LatexCtx(ctx))
		== "(\\neg (x:5 > 100 \\rightarrow \\text{False}) \\rightarrow \\text{True})"
	)


def test_latex_with_context_special_expressions() -> None:
	"""Test LaTeX output of special expressions with context."""
	x = Var[float, Ctx]("f")  # Using f=1.5 from ctx

	# PlusMinus (evaluates to itself)
	pm = PlusMinus("measurement", 5.0, 0.1)
	assert latex(pm) == "measurement \\pm 0.1"
	assert latex(pm, LatexCtx(ctx)) == "(measurement \\pm 0.1 \\rightarrow measurement \\pm 0.1)"

	# Percent (evaluates to itself)
	pct = Percent("tolerance", 100.0, 5.0)
	assert latex(pct) == "tolerance \\pm 5.0\\%"
	assert latex(pct, LatexCtx(ctx)) == "(tolerance \\pm 5.0\\% \\rightarrow tolerance \\pm 5.0\\%)"

	# Approximately
	target = PlusMinus("target", 1.5, 0.1)
	approx = Approximately(x, target)
	assert latex(approx) == "f \\approx target \\pm 0.1"
	assert (
		latex(approx, LatexCtx(ctx)) == "(f:1.5 \\approx target \\pm 0.1 \\rightarrow \\text{True})"
	)

	# Predicate
	pred = Predicate("condition", x > 1.0)
	assert latex(pred) == "\\text{condition}: f > 1.0"
	assert (
		latex(pred, LatexCtx(ctx))
		== "(\\text{condition}: (f:1.5 > 1.0 \\rightarrow \\text{True}) \\rightarrow \\text{True})"
	)


def test_latex_with_context_complex_expressions() -> None:
	"""Test LaTeX output of complex expressions with context."""
	x = Var[float, Ctx]("f")  # f=1.5
	y = Var[int, Ctx]("x")  # x=5
	z = Var[int, Ctx]("y")  # y=10

	# Nested arithmetic with proper parentheses handling
	complex_expr = (x + y) * z
	assert latex(complex_expr) == "(f + x) \\cdot y"
	assert (
		latex(complex_expr, LatexCtx(ctx))
		== "(((f:1.5 + x:5 \\rightarrow 6.5)) \\cdot y:10 \\rightarrow 65.0)"
	)

	# Polynomial-like expression
	poly_expr = y**2 + y + 1
	assert latex(poly_expr) == "x^2 + x + 1"
	assert (
		latex(poly_expr, LatexCtx(ctx))
		== "(((x:5^2 \\rightarrow 25) + x:5 \\rightarrow 30) + 1 \\rightarrow 31)"
	)


def test_latex_with_context_edge_cases() -> None:
	"""Test LaTeX output with context for edge cases."""
	x = Var[int, Ctx]("x")

	# Zero result
	zero_expr = x - x
	assert latex(zero_expr) == "x - x"
	assert latex(zero_expr, LatexCtx(ctx)) == "(x:5 - x:5 \\rightarrow 0)"

	# Negative values
	neg_expr = x - 10
	assert latex(neg_expr) == "x - 10"
	assert latex(neg_expr, LatexCtx(ctx)) == "(x:5 - 10 \\rightarrow -5)"

	# Float values
	x_float = Var[float, Ctx]("f")  # f=1.5
	float_expr = x_float * 2
	assert latex(float_expr) == "f \\cdot 2"
	assert latex(float_expr, LatexCtx(ctx)) == "(f:1.5 \\cdot 2 \\rightarrow 3.0)"


def test_latex_boolean_expressions_with_context() -> None:
	"""Test LaTeX output of expressions that evaluate to booleans."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")

	# Boolean expressions should show their values as \\text{True/False} when evaluated
	true_expr = x < y  # 5 < 10 is True
	false_expr = x > y  # 5 > 10 is False

	# Without context - shows structure
	assert latex(true_expr) == "x < y"
	assert latex(false_expr) == "x > y"

	# With context - shows expression -> boolean result
	assert latex(true_expr, LatexCtx(ctx)) == "(x:5 < y:10 \\rightarrow \\text{True})"
	assert latex(false_expr, LatexCtx(ctx)) == "(x:5 > y:10 \\rightarrow \\text{False})"


# Show Flags Tests


def test_show_flags_no_flags() -> None:
	"""Test LaTeX output with no Show flags."""

	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = x + y

	# No flags - shows structure with result
	assert latex(expr, LatexCtx(ctx, Show(0))) == "(x + y \\rightarrow 15)"


def test_show_flags_values_only() -> None:
	"""Test LaTeX output with VALUES flag only."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = x + y

	# VALUES only - shows values with final result wrapper
	assert latex(expr, LatexCtx(ctx, Show.VALUES)) == "(x:5 + y:10 \\rightarrow 15)"


def test_show_flags_work_only() -> None:
	"""Test LaTeX output with WORK flag only."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = x + y

	# WORK only - shows structure and result
	assert latex(expr, LatexCtx(ctx, Show.WORK)) == "(x + y \\rightarrow 15)"


def test_show_flags_values_and_work() -> None:
	"""Test LaTeX output with both VALUES and WORK flags."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = x + y

	# Both VALUES and WORK - shows values and result
	assert latex(expr, LatexCtx(ctx, Show.VALUES | Show.WORK)) == "(x:5 + y:10 \\rightarrow 15)"


def test_show_flags_complex_expressions() -> None:
	"""Test Show flags with complex expressions."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = (x + y) * 2

	# Structure only - shows structure with result
	assert latex(expr, LatexCtx(ctx, Show(0))) == "((x + y) \\cdot 2 \\rightarrow 30)"

	# VALUES only
	assert latex(expr, LatexCtx(ctx, Show.VALUES)) == "((x:5 + y:10) \\cdot 2 \\rightarrow 30)"

	# WORK only
	assert (
		latex(expr, LatexCtx(ctx, Show.WORK))
		== "(((x + y \\rightarrow 15)) \\cdot 2 \\rightarrow 30)"
	)

	# Both FLAGS
	assert (
		latex(expr, LatexCtx(ctx, Show.VALUES | Show.WORK))
		== "(((x:5 + y:10 \\rightarrow 15)) \\cdot 2 \\rightarrow 30)"
	)


def test_show_flags_boolean_expressions() -> None:
	"""Test Show flags with boolean expressions."""
	x = Var[int, Ctx]("x")
	y = Var[int, Ctx]("y")
	expr = x < y

	# Structure only - shows structure with result
	assert latex(expr, LatexCtx(ctx, Show(0))) == "(x < y \\rightarrow \\text{True})"

	# VALUES only - shows values with result
	assert latex(expr, LatexCtx(ctx, Show.VALUES)) == "(x:5 < y:10 \\rightarrow \\text{True})"

	# WORK only
	assert latex(expr, LatexCtx(ctx, Show.WORK)) == "(x < y \\rightarrow \\text{True})"

	# Both FLAGS
	assert (
		latex(expr, LatexCtx(ctx, Show.VALUES | Show.WORK))
		== "(x:5 < y:10 \\rightarrow \\text{True})"
	)


def test_show_flags_with_constants() -> None:
	"""Test Show flags with named constants."""
	x = Var[int, Ctx]("x")
	Max = Const("Max", 20)
	expr = x + Max

	# Structure only shows constant name with result
	assert latex(expr, LatexCtx(ctx, Show(0))) == "(x + Max \\rightarrow 25)"

	# VALUES shows constant name:value with result
	assert latex(expr, LatexCtx(ctx, Show.VALUES)) == "(x:5 + Max:20 \\rightarrow 25)"

	# WORK shows structure and result
	assert latex(expr, LatexCtx(ctx, Show.WORK)) == "(x + Max \\rightarrow 25)"

	# Both show values and result
	assert latex(expr, LatexCtx(ctx, Show.VALUES | Show.WORK)) == "(x:5 + Max:20 \\rightarrow 25)"


# Statistical Operations Tests


def test_mean_latex() -> None:
	"""Test mean operation LaTeX formatting."""
	values = Var[SizedIterable[float], StatsCtx]("values")
	mean_expr = Mean(values)

	# Without context - shows mathematical notation
	assert latex(mean_expr) == "\\bar{values}"

	# With context - shows values and result
	assert latex(mean_expr, LatexCtx(stats_ctx)) == "(\\bar{values:5[1.0,..5.0]} \\rightarrow 3.0)"


def test_stddev_latex() -> None:
	"""Test standard deviation LaTeX formatting."""
	values = Var[SizedIterable[float], StatsCtx]("values")
	stddev_expr = StdDev(values)

	# Without context - shows mathematical notation
	assert latex(stddev_expr) == "\\sigma_{values}"

	# With context - shows values and result
	result_str = latex(stddev_expr, LatexCtx(stats_ctx))
	assert result_str.startswith("(\\sigma_{values:5[1.0,..5.0]} \\rightarrow")
	assert "1.58" in result_str  # Standard deviation of [1,2,3,4,5]


def test_median_latex() -> None:
	"""Test median LaTeX formatting."""
	values = Var[SizedIterable[float], StatsCtx]("values")
	median_expr = Median(values)

	# Without context - shows mathematical notation
	assert latex(median_expr) == "\\text{median}(values)"

	# With context - shows values and result
	assert (
		latex(median_expr, LatexCtx(stats_ctx))
		== "(\\text{median}(values:5[1.0,..5.0]) \\rightarrow 3.0)"
	)


def test_percentile_latex() -> None:
	"""Test percentile LaTeX formatting."""
	values = Var[SizedIterable[float], StatsCtx]("values")
	p95_expr = Percentile(95, values)
	p50_expr = Percentile(50.5, values)

	# Without context - shows mathematical notation
	assert latex(p95_expr) == "P_{95}(values)"
	assert latex(p50_expr) == "P_{50.5}(values)"

	# With context - shows values and result
	assert latex(p95_expr, LatexCtx(stats_ctx)) == "(P_{95}(values:5[1.0,..5.0]) \\rightarrow 4.8)"


def test_range_latex() -> None:
	"""Test range LaTeX formatting."""
	values = Var[SizedIterable[float], StatsCtx]("values")
	range_expr = Range(values)

	# Without context - shows mathematical notation
	assert latex(range_expr) == "\\text{range}(values)"

	# With context - shows values and result
	assert (
		latex(range_expr, LatexCtx(stats_ctx))
		== "(\\text{range}(values:5[1.0,..5.0]) \\rightarrow 4.0)"
	)


def test_count_latex() -> None:
	"""Test count LaTeX formatting."""
	values = Var[SizedIterable[float], StatsCtx]("values")
	count_expr = Count(values)

	# Without context - shows mathematical notation
	assert latex(count_expr) == "|values|"

	# With context - shows values and result
	assert latex(count_expr, LatexCtx(stats_ctx)) == "(|values:5[1.0,..5.0]| \\rightarrow 5)"


def test_statistical_operations_with_show_flags() -> None:
	"""Test statistical operations with different Show flags."""
	values = Var[SizedIterable[float], StatsCtx]("values")
	mean_expr = Mean(values)

	# No flags - shows structure with result
	assert latex(mean_expr, LatexCtx(stats_ctx, Show(0))) == "(\\bar{values} \\rightarrow 3.0)"

	# VALUES only - shows values with result
	assert (
		latex(mean_expr, LatexCtx(stats_ctx, Show.VALUES))
		== "(\\bar{values:5[1.0,..5.0]} \\rightarrow 3.0)"
	)

	# WORK only - shows structure with result
	assert latex(mean_expr, LatexCtx(stats_ctx, Show.WORK)) == "(\\bar{values} \\rightarrow 3.0)"

	# Both flags - shows values with result
	assert (
		latex(mean_expr, LatexCtx(stats_ctx, Show.VALUES | Show.WORK))
		== "(\\bar{values:5[1.0,..5.0]} \\rightarrow 3.0)"
	)


def test_statistical_operations_with_greek_variables() -> None:
	"""Test statistical operations with Greek letter variable names."""

	# Create a context with Greek variable names
	class GreekCtx(NamedTuple):
		sigma: SizedIterable[float] = [1.0, 2.0, 3.0, 4.0, 5.0]
		alpha: SizedIterable[float] = [1.0, 2.0, 3.0, 4.0, 5.0]

	greek_ctx = GreekCtx()
	sigma_data = Var[SizedIterable[float], GreekCtx]("sigma")
	mean_expr = Mean(sigma_data)

	# Variable name should be converted to Greek letter
	assert latex(mean_expr) == "\\bar{\\sigma}"
	assert (
		latex(mean_expr, LatexCtx(greek_ctx, Show.VALUES))
		== "(\\bar{\\sigma:5[1.0,..5.0]} \\rightarrow 3.0)"
	)


def test_complex_statistical_expressions() -> None:
	"""Test complex expressions involving statistical operations."""
	values = Var[SizedIterable[float], StatsCtx]("values")
	mean_expr = Mean(values)
	stddev_expr = StdDev(values)

	# Statistical operations in arithmetic
	range_normalized = (mean_expr - 1) / stddev_expr

	# Without context
	assert latex(range_normalized) == "\\frac{\\bar{values} - 1}{\\sigma_{values}}"

	# With context (showing intermediate results)
	result_str = latex(range_normalized, LatexCtx(stats_ctx, Show.VALUES | Show.WORK))
	assert "\\bar{values:5[1.0,..5.0]}" in result_str
	assert "\\sigma_{values:5[1.0,..5.0]}" in result_str


def test_statistical_operations_in_comparisons() -> None:
	"""Test statistical operations used in comparisons."""
	values = Var[SizedIterable[float], StatsCtx]("values")
	mean_expr = Mean(values)
	target = 3.0

	# Comparison with statistical operation
	quality_check = mean_expr > target

	assert latex(quality_check) == "\\bar{values} > 3.0"
	assert (
		latex(quality_check, LatexCtx(stats_ctx, Show.VALUES))
		== "(\\bar{values:5[1.0,..5.0]} > 3.0 \\rightarrow \\text{False})"
	)


def test_statistical_operations_with_subscripts() -> None:
	"""Test statistical operations with subscripted variable names."""
	x_measurements = Var[SizedIterable[float], StatsCtx]("measurements")
	mean_expr = Mean(x_measurements)

	# Test with measurements variable
	assert latex(mean_expr) == "\\bar{measurements}"

	# With context using measurements data
	measurements_ctx = StatsCtx(measurements=[10.0, 15.0, 20.0, 25.0, 30.0])
	assert (
		latex(mean_expr, LatexCtx(measurements_ctx, Show.VALUES))
		== "(\\bar{measurements:5[10.0,..30.0]} \\rightarrow 20.0)"
	)
